# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Main entry point into the Identity service."""

import copy
import functools
import itertools
import operator
import os
import threading
import uuid

from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils
from pycadf import reason
import stevedore

from keystone import assignment  # TODO(lbragstad): Decouple this dependency
from keystone.common import cache
from keystone.common import driver_hints
from keystone.common import manager
from keystone.common import provider_api
from keystone.common.validation import validators
import keystone.conf
from keystone import exception
from keystone.i18n import _
from keystone.identity.mapping_backends import mapping
from keystone import notifications

CONF = keystone.conf.CONF

LOG = log.getLogger(__name__)

PROVIDERS = provider_api.ProviderAPIs

MEMOIZE = cache.get_memoization_decorator(group='identity')

ID_MAPPING_REGION = cache.create_region(name='id mapping')
MEMOIZE_ID_MAPPING = cache.get_memoization_decorator(
    group='identity', region=ID_MAPPING_REGION
)

DOMAIN_CONF_FHEAD = 'keystone.'
DOMAIN_CONF_FTAIL = '.conf'

# The number of times we will attempt to register a domain to use the SQL
# driver, if we find that another process is in the middle of registering or
# releasing at the same time as us.
REGISTRATION_ATTEMPTS = 10

# Config Registration Types
SQL_DRIVER = 'SQL'


def get_driver(namespace, driver_name, *args):
    """Get identity driver without initializing.

    The method is invoked to be able to introspect domain specific driver
    looking for additional configuration options required by the driver.
    """
    try:
        driver_manager = stevedore.DriverManager(
            namespace, driver_name, invoke_on_load=False, invoke_args=args
        )
        return driver_manager.driver
    except stevedore.exception.NoMatches:
        msg = _('Unable to find %(name)r driver in %(namespace)r.')
        raise ImportError(msg % {'name': driver_name, 'namespace': namespace})


class DomainConfigs(provider_api.ProviderAPIMixin, dict):
    """Discover, store and provide access to domain specific configs.

    The setup_domain_drivers() call will be made via the wrapper from
    the first call to any driver function handled by this manager.

    Domain specific configurations are only supported for the identity backend
    and the individual configurations are either specified in the resource
    database or in individual domain configuration files, depending on the
    setting of the 'domain_configurations_from_database' config option.

    The result will be that for each domain with a specific configuration,
    this class will hold a reference to a ConfigOpts and driver object that
    the identity manager and driver can use.

    """

    configured = False
    driver = None
    _any_sql = False
    lock = threading.Lock()

    def _load_driver(self, domain_config):
        return manager.load_driver(
            Manager.driver_namespace,
            domain_config['cfg'].identity.driver,
            domain_config['cfg'],
        )

    def _load_config_from_file(self, resource_api, file_list, domain_name):
        def _assert_no_more_than_one_sql_driver(new_config, config_file):
            """Ensure there is no more than one sql driver.

            Check to see if the addition of the driver in this new config
            would cause there to be more than one sql driver.

            """
            if new_config['driver'].is_sql and (
                self.driver.is_sql or self._any_sql
            ):
                # The addition of this driver would cause us to have more than
                # one sql driver, so raise an exception.
                raise exception.MultipleSQLDriversInConfig(source=config_file)
            self._any_sql = self._any_sql or new_config['driver'].is_sql

        try:
            domain_ref = resource_api.get_domain_by_name(domain_name)
        except exception.DomainNotFound:
            LOG.warning(
                'Invalid domain name (%s) found in config file name',
                domain_name,
            )
            return

        # Create a new entry in the domain config dict, which contains
        # a new instance of both the conf environment and driver using
        # options defined in this set of config files.  Later, when we
        # service calls via this Manager, we'll index via this domain
        # config dict to make sure we call the right driver
        domain_config = {}
        domain_config['cfg'] = cfg.ConfigOpts()
        keystone.conf.configure(conf=domain_config['cfg'])
        domain_config['cfg'](
            args=[],
            project='keystone',
            default_config_files=file_list,
            default_config_dirs=[],
        )
        domain_config['driver'] = self._load_driver(domain_config)
        _assert_no_more_than_one_sql_driver(domain_config, file_list)
        self[domain_ref['id']] = domain_config

    def _setup_domain_drivers_from_files(self, standard_driver, resource_api):
        """Read the domain specific configuration files and load the drivers.

        Domain configuration files are stored in the domain config directory,
        and must be named of the form:

        keystone.<domain_name>.conf

        For each file, call the load config method where the domain_name
        will be turned into a domain_id and then:

        - Create a new config structure, adding in the specific additional
          options defined in this config file
        - Initialise a new instance of the required driver with this new config

        """
        conf_dir = CONF.identity.domain_config_dir
        if not os.path.exists(conf_dir):
            LOG.warning(
                'Unable to locate domain config directory: %s', conf_dir
            )
            return

        for r, d, f in os.walk(conf_dir):
            for fname in f:
                if fname.startswith(DOMAIN_CONF_FHEAD) and fname.endswith(
                    DOMAIN_CONF_FTAIL
                ):
                    if fname.count('.') >= 2:
                        self._load_config_from_file(
                            resource_api,
                            [os.path.join(r, fname)],
                            fname[
                                len(DOMAIN_CONF_FHEAD) : -len(
                                    DOMAIN_CONF_FTAIL
                                )
                            ],
                        )
                    else:
                        LOG.debug(
                            (
                                'Ignoring file (%s) while scanning domain '
                                'config directory'
                            ),
                            fname,
                        )

    def _load_config_from_database(self, domain_id, specific_config):
        def _assert_no_more_than_one_sql_driver(domain_id, new_config):
            """Ensure adding driver doesn't push us over the limit of 1.

            The checks we make in this method need to take into account that
            we may be in a multiple process configuration and ensure that
            any race conditions are avoided.

            """
            if not new_config['driver'].is_sql:
                PROVIDERS.domain_config_api.release_registration(domain_id)
                return

            # To ensure the current domain is the only SQL driver, we attempt
            # to register our use of SQL. If we get it we know we are good,
            # if we fail to register it then we should:
            #
            # - First check if another process has registered for SQL for our
            #   domain, in which case we are fine
            # - If a different domain has it, we should check that this domain
            #   is still valid, in case, for example, domain deletion somehow
            #   failed to remove its registration (i.e. we self heal for these
            #   kinds of issues).

            domain_registered = 'Unknown'
            for attempt in range(REGISTRATION_ATTEMPTS):
                if PROVIDERS.domain_config_api.obtain_registration(
                    domain_id, SQL_DRIVER
                ):
                    LOG.debug(
                        'Domain %s successfully registered to use the '
                        'SQL driver.',
                        domain_id,
                    )
                    return

                # We failed to register our use, let's find out who is using it
                try:
                    domain_registered = (
                        PROVIDERS.domain_config_api.read_registration(
                            SQL_DRIVER
                        )
                    )
                except exception.ConfigRegistrationNotFound:
                    msg = (
                        'While attempting to register domain %(domain)s to '
                        'use the SQL driver, another process released it, '
                        'retrying (attempt %(attempt)s).'
                    )
                    LOG.debug(
                        msg, {'domain': domain_id, 'attempt': attempt + 1}
                    )
                    continue

                if domain_registered == domain_id:
                    # Another process already registered it for us, so we are
                    # fine. In the race condition when another process is
                    # in the middle of deleting this domain, we know the domain
                    # is already disabled and hence telling the caller that we
                    # are registered is benign.
                    LOG.debug(
                        'While attempting to register domain %s to use '
                        'the SQL driver, found that another process had '
                        'already registered this domain. This is normal '
                        'in multi-process configurations.',
                        domain_id,
                    )
                    return

                # So we don't have it, but someone else does...let's check that
                # this domain is still valid
                try:
                    PROVIDERS.resource_api.get_domain(domain_registered)
                except exception.DomainNotFound:
                    msg = (
                        'While attempting to register domain %(domain)s to '
                        'use the SQL driver, found that it was already '
                        'registered to a domain that no longer exists '
                        '(%(old_domain)s). Removing this stale '
                        'registration and retrying (attempt %(attempt)s).'
                    )
                    LOG.debug(
                        msg,
                        {
                            'domain': domain_id,
                            'old_domain': domain_registered,
                            'attempt': attempt + 1,
                        },
                    )
                    PROVIDERS.domain_config_api.release_registration(
                        domain_registered, type=SQL_DRIVER
                    )
                    continue

                # The domain is valid, so we really do have an attempt at more
                # than one SQL driver.
                details = (
                    _('Config API entity at /domains/%s/config') % domain_id
                )
                raise exception.MultipleSQLDriversInConfig(source=details)

            # We fell out of the loop without either registering our domain or
            # being able to find who has it...either we were very very very
            # unlucky or something is awry.
            msg = _(
                'Exceeded attempts to register domain %(domain)s to use '
                'the SQL driver, the last domain that appears to have '
                'had it is %(last_domain)s, giving up'
            ) % {'domain': domain_id, 'last_domain': domain_registered}
            raise exception.UnexpectedError(msg)

        domain_config = {}
        domain_config['cfg'] = cfg.ConfigOpts()
        keystone.conf.configure(conf=domain_config['cfg'])
        domain_config['cfg'](
            args=[],
            project='keystone',
            default_config_files=[],
            default_config_dirs=[],
        )

        # Try to identify the required driver for the domain to let it register
        # supported configuration options. In difference to the FS based
        # configuration this is being set through `oslo_cfg.set_override` and
        # thus require special treatment.
        try:
            driver_name = specific_config.get("identity", {}).get(
                "driver", domain_config["cfg"].identity.driver
            )
            # For the non in-tree drivers ...
            if driver_name not in ["sql", "ldap"]:
                # Locate the driver without invoking ...
                driver = get_driver(Manager.driver_namespace, driver_name)
                # Check whether it wants to register additional config options
                # ...
                if hasattr(driver, "register_opts"):
                    # And register them for the domain_config (not the global
                    # Keystone config)
                    driver.register_opts(domain_config["cfg"])
        except Exception as ex:
            # If we failed for some reason - something wrong with the driver,
            # so let's just skip registering config options. This matches older
            # behavior of Keystone where out-of-tree drivers were not able to
            # register config options with the DB configuration loading branch.
            LOG.debug(
                f"Exception during attempt to load domain specific "
                f"configuration options: {ex}"
            )

        # Override any options that have been passed in as specified in the
        # database.
        for group in specific_config:
            for option in specific_config[group]:
                # NOTE(gtema): Very first time default driver is being ordered
                # to process the domain. This will change once initialization
                # completes. Until the driver specific configuration is being
                # registered `set_override` will fail for options not known by
                # the core Keystone. Make this loading not failing letting code
                # to complete the process properly.
                try:
                    domain_config['cfg'].set_override(
                        option, specific_config[group][option], group
                    )
                except (cfg.NoSuchOptError, cfg.NoSuchGroupError):
                    # Error to register config overrides for wrong driver. This
                    # is not worth of logging since it is a normal case during
                    # Keystone initialization.
                    pass

        domain_config['cfg_overrides'] = specific_config
        domain_config['driver'] = self._load_driver(domain_config)
        _assert_no_more_than_one_sql_driver(domain_id, domain_config)
        self[domain_id] = domain_config

    def _setup_domain_drivers_from_database(
        self, standard_driver, resource_api
    ):
        """Read domain specific configuration from database and load drivers.

        Domain configurations are stored in the domain-config backend,
        so we go through each domain to find those that have a specific config
        defined, and for those that do we:

        - Create a new config structure, overriding any specific options
          defined in the resource backend
        - Initialise a new instance of the required driver with this new config

        """
        for domain in resource_api.list_domains():
            domain_config_options = (
                PROVIDERS.domain_config_api.get_config_with_sensitive_info(
                    domain['id']
                )
            )
            if domain_config_options:
                self._load_config_from_database(
                    domain['id'], domain_config_options
                )

    def setup_domain_drivers(self, standard_driver, resource_api):
        # This is called by the api call wrapper
        self.driver = standard_driver

        if CONF.identity.domain_configurations_from_database:
            self._setup_domain_drivers_from_database(
                standard_driver, resource_api
            )
        else:
            self._setup_domain_drivers_from_files(
                standard_driver, resource_api
            )
        self.configured = True

    def get_domain_driver(self, domain_id):
        self.check_config_and_reload_domain_driver_if_required(domain_id)
        if domain_id in self:
            return self[domain_id]['driver']

    def get_domain_conf(self, domain_id):
        self.check_config_and_reload_domain_driver_if_required(domain_id)
        if domain_id in self:
            return self[domain_id]['cfg']
        else:
            return CONF

    def reload_domain_driver(self, domain_id):
        # Only used to support unit tests that want to set
        # new config values.  This should only be called once
        # the domains have been configured, since it relies on
        # the fact that the configuration files/database have already been
        # read.
        if self.configured:
            if domain_id in self:
                self[domain_id]['driver'] = self._load_driver(self[domain_id])
            else:
                # The standard driver
                self.driver = self.driver()

    def check_config_and_reload_domain_driver_if_required(self, domain_id):
        """Check for, and load, any new domain specific config for this domain.

        This is only supported for the database-stored domain specific
        configuration.

        When the domain specific drivers were set up, we stored away the
        specific config for this domain that was available at that time. So we
        now read the current version and compare. While this might seem
        somewhat inefficient, the sensitive config call is cached, so should be
        light weight. More importantly, when the cache timeout is reached, we
        will get any config that has been updated from any other keystone
        process.

        This cache-timeout approach works for both multi-process and
        multi-threaded keystone configurations. In multi-threaded
        configurations, even though we might remove a driver object (that
        could be in use by another thread), this won't actually be thrown away
        until all references to it have been broken. When that other
        thread is released back and is restarted with another command to
        process, next time it accesses the driver it will pickup the new one.

        """
        if (
            not CONF.identity.domain_specific_drivers_enabled
            or not CONF.identity.domain_configurations_from_database
        ):
            # If specific drivers are not enabled, then there is nothing to do.
            # If we are not storing the configurations in the database, then
            # we'll only re-read the domain specific config files on startup
            # of keystone.
            return

        latest_domain_config = (
            PROVIDERS.domain_config_api.get_config_with_sensitive_info(
                domain_id
            )
        )
        domain_config_in_use = domain_id in self

        if latest_domain_config:
            if (
                not domain_config_in_use
                or latest_domain_config != self[domain_id]['cfg_overrides']
            ):
                self._load_config_from_database(
                    domain_id, latest_domain_config
                )
        elif domain_config_in_use:
            # The domain specific config has been deleted, so should remove the
            # specific driver for this domain.
            try:
                del self[domain_id]
            except KeyError:  # nosec
                # Allow this error in case we are unlucky and in a
                # multi-threaded situation, two threads happen to be running
                # in lock step.
                pass
        # If we fall into the else condition, this means there is no domain
        # config set, and there is none in use either, so we have nothing
        # to do.


def domains_configured(f):
    """Wrap API calls to lazy load domain configs after init.

    This is required since the assignment manager needs to be initialized
    before this manager, and yet this manager's init wants to be
    able to make assignment calls (to build the domain configs).  So
    instead, we check if the domains have been initialized on entry
    to each call, and if requires load them,

    """

    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        if (
            not self.domain_configs.configured
            and CONF.identity.domain_specific_drivers_enabled
        ):
            # If domain specific driver has not been configured, acquire the
            # lock and proceed with loading the driver.
            with self.domain_configs.lock:
                # Check again just in case some other thread has already
                # completed domain config.
                if not self.domain_configs.configured:
                    self.domain_configs.setup_domain_drivers(
                        self.driver, PROVIDERS.resource_api
                    )
        return f(self, *args, **kwargs)

    return wrapper


def exception_translated(exception_type):
    """Wrap API calls to map to correct exception."""

    def _exception_translated(f):
        @functools.wraps(f)
        def wrapper(self, *args, **kwargs):
            try:
                return f(self, *args, **kwargs)
            except exception.PublicIDNotFound as e:
                if exception_type == 'user':
                    raise exception.UserNotFound(user_id=str(e))
                elif exception_type == 'group':
                    raise exception.GroupNotFound(group_id=str(e))
                elif exception_type == 'assertion':
                    raise AssertionError(_('Invalid user / password'))
                else:
                    raise

        return wrapper

    return _exception_translated


@notifications.listener
class Manager(manager.Manager):
    """Default pivot point for the Identity backend.

    See :mod:`keystone.common.manager.Manager` for more details on how this
    dynamically calls the backend.

    This class also handles the support of domain specific backends, by using
    the DomainConfigs class. The setup call for DomainConfigs is called
    from with the @domains_configured wrapper in a lazy loading fashion
    to get around the fact that we can't satisfy the assignment api it needs
    from within our __init__() function since the assignment driver is not
    itself yet initialized.

    Each of the identity calls are pre-processed here to choose, based on
    domain, which of the drivers should be called. The non-domain-specific
    driver is still in place, and is used if there is no specific driver for
    the domain in question (or we are not using multiple domain drivers).

    Starting with Juno, in order to be able to obtain the domain from
    just an ID being presented as part of an API call, a public ID to domain
    and local ID mapping is maintained.  This mapping also allows for the local
    ID of drivers that do not provide simple UUIDs (such as LDAP) to be
    referenced via a public facing ID.  The mapping itself is automatically
    generated as entities are accessed via the driver.

    This mapping is only used when:
    - the entity is being handled by anything other than the default driver, or
    - the entity is being handled by the default LDAP driver and backward
    compatible IDs are not required.

    This means that in the standard case of a single SQL backend or the default
    settings of a single LDAP backend (since backward compatible IDs is set to
    True by default), no mapping is used. An alternative approach would be to
    always use the mapping table, but in the cases where we don't need it to
    make the public and local IDs the same. It is felt that not using the
    mapping by default is a more prudent way to introduce this functionality.

    """

    driver_namespace = 'keystone.identity'
    _provides_api = 'identity_api'

    _USER = 'user'
    _GROUP = 'group'

    def __init__(self):
        super().__init__(CONF.identity.driver)
        self.domain_configs = DomainConfigs()
        notifications.register_event_callback(
            notifications.ACTIONS.internal,
            notifications.DOMAIN_DELETED,
            self._domain_deleted,
        )
        self.event_callbacks = {
            notifications.ACTIONS.deleted: {
                'project': [self._unset_default_project]
            }
        }

    def _domain_deleted(self, service, resource_type, operation, payload):
        domain_id = payload['resource_info']

        driver = self._select_identity_driver(domain_id)

        if driver.is_sql:
            group_refs = self.list_groups(domain_scope=domain_id)
            for group in group_refs:
                # Cleanup any existing groups.
                try:
                    self.delete_group(group['id'])
                except exception.GroupNotFound:
                    LOG.debug(
                        (
                            'Group %(groupid)s not found when deleting '
                            'domain contents for %(domainid)s, continuing '
                            'with cleanup.'
                        ),
                        {'groupid': group['id'], 'domainid': domain_id},
                    )

        # And finally, delete the users themselves
        user_refs = self.list_users(domain_scope=domain_id)

        for user in user_refs:
            try:
                if not driver.is_sql:
                    PROVIDERS.shadow_users_api.delete_user(user['id'])
                else:
                    self.delete_user(user['id'])
            except exception.UserNotFound:
                LOG.debug(
                    (
                        'User %(userid)s not found when deleting domain '
                        'contents for %(domainid)s, continuing with '
                        'cleanup.'
                    ),
                    {'userid': user['id'], 'domainid': domain_id},
                )

    def _unset_default_project(
        self, service, resource_type, operation, payload
    ):
        """Callback, clears user default_project_id after project deletion.

        Notifications are used to unset a user's default project because
        there is no foreign key to the project. Projects can be in a non-SQL
        backend, making FKs impossible.

        """
        project_id = payload['resource_info']
        drivers = itertools.chain(
            self.domain_configs.values(), [{'driver': self.driver}]
        )
        for d in drivers:
            try:
                d['driver'].unset_default_project_id(project_id)
            except exception.Forbidden:
                # NOTE(lbragstad): If the driver throws a Forbidden, it's
                # because the driver doesn't support writes. This is the case
                # with the in-tree LDAP implementation since it is read-only.
                # This also ensures consistency for out-of-tree backends that
                # might be read-only.
                pass

    # Domain ID normalization methods
    def _set_domain_id_and_mapping(self, ref, domain_id, driver, entity_type):
        """Patch the domain_id/public_id into the resulting entity(ies).

        :param ref: the entity or list of entities to post process
        :param domain_id: the domain scope used for the call
        :param driver: the driver used to execute the call
        :param entity_type: whether this is a user or group

        :returns: post processed entity or list or entities

        Called to post-process the entity being returned, using a mapping
        to substitute a public facing ID as necessary. This method must
        take into account:

        - If the driver is not domain aware, then we must set the domain
          attribute of all entities irrespective of mapping.
        - If the driver does not support UUIDs, then we always want to provide
          a mapping, except for the special case of this being the default
          driver and backward_compatible_ids is set to True. This is to ensure
          that entity IDs do not change for an existing LDAP installation (only
          single domain/driver LDAP configurations were previously supported).
        - If the driver does support UUIDs, then we always create a mapping
          entry, but use the local UUID as the public ID.  The exception to
          this is that if we just have single driver (i.e. not using specific
          multi-domain configs), then we don't bother with the mapping at all.

        """
        conf = CONF.identity

        if not self._needs_post_processing(driver):
            # a classic case would be when running with a single SQL driver
            return ref

        LOG.debug(
            'ID Mapping - Domain ID: %(domain)s, '
            'Default Driver: %(driver)s, '
            'Domains: %(aware)s, UUIDs: %(generate)s, '
            'Compatible IDs: %(compat)s',
            {
                'domain': domain_id,
                'driver': (driver == self.driver),
                'aware': driver.is_domain_aware(),
                'generate': driver.generates_uuids(),
                'compat': CONF.identity_mapping.backward_compatible_ids,
            },
        )

        if isinstance(ref, dict):
            return self._set_domain_id_and_mapping_for_single_ref(
                ref, domain_id, driver, entity_type, conf
            )
        elif isinstance(ref, list):
            return self._set_domain_id_and_mapping_for_list(
                ref, domain_id, driver, entity_type, conf
            )
        else:
            raise ValueError(_('Expected dict or list: %s') % type(ref))

    def _needs_post_processing(self, driver):
        """Return whether entity from driver needs domain added or mapping."""
        return (
            driver is not self.driver
            or not driver.generates_uuids()
            or not driver.is_domain_aware()
        )

    def _insert_new_public_id(self, local_entity, ref, driver):
        # Need to create a mapping. If the driver generates UUIDs
        # then pass the local UUID in as the public ID to use.
        public_id = None
        if driver.generates_uuids():
            public_id = ref['id']
        ref['id'] = PROVIDERS.id_mapping_api.create_id_mapping(
            local_entity, public_id
        )
        LOG.debug('Created new mapping to public ID: %s', ref['id'])

    def _set_domain_id_and_mapping_for_single_ref(
        self, ref, domain_id, driver, entity_type, conf
    ):
        LOG.debug('Local ID: %s', ref['id'])
        ref = ref.copy()

        if not driver.is_domain_aware():
            if not domain_id:
                domain_id = CONF.identity.default_domain_id
            ref['domain_id'] = domain_id

        if self._is_mapping_needed(driver):
            local_entity = {
                'domain_id': ref['domain_id'],
                'local_id': ref['id'],
                'entity_type': entity_type,
            }
            public_id = PROVIDERS.id_mapping_api.get_public_id(local_entity)
            if public_id:
                ref['id'] = public_id
                LOG.debug('Found existing mapping to public ID: %s', ref['id'])
            else:
                self._insert_new_public_id(local_entity, ref, driver)
        return ref

    def _set_domain_id_and_mapping_for_list(
        self, ref_list, domain_id, driver, entity_type, conf
    ):
        """Set domain id and mapping for a list of refs.

        The method modifies refs in-place.
        """
        if not ref_list:
            return []

        # If the domain_id is None that means we are running in a single
        # backend mode, so to remain backwards compatible we will use the
        # default domain ID.
        if not domain_id:
            domain_id = CONF.identity.default_domain_id

        if not driver.is_domain_aware():
            for ref in ref_list:
                ref['domain_id'] = domain_id

        if not self._is_mapping_needed(driver):
            return ref_list

        # build a map of refs for fast look-up
        refs_map = {}
        for r in ref_list:
            refs_map[(r['id'], entity_type, r['domain_id'])] = r

        # fetch all mappings for the domain, lookup the user at the map built
        # at previous step and replace his id.
        domain_mappings = PROVIDERS.id_mapping_api.get_domain_mapping_list(
            domain_id, entity_type=entity_type
        )
        for _mapping in domain_mappings:
            idx = (_mapping.local_id, _mapping.entity_type, _mapping.domain_id)
            try:
                ref = refs_map.pop(idx)
                # due to python specifics, `ref` still points to an item in
                # `ref_list`. That's why when we change it here, it gets
                # changed in `ref_list`.
                ref['id'] = _mapping.public_id
            except KeyError:
                pass  # some old entry, skip it

        # at this point, all known refs were granted a public_id. For the refs
        # left, there are no mappings. They need to be created.
        for ref in refs_map.values():
            local_entity = {
                'domain_id': ref['domain_id'],
                'local_id': ref['id'],
                'entity_type': entity_type,
            }
            self._insert_new_public_id(local_entity, ref, driver)
        return ref_list

    def _is_mapping_needed(self, driver):
        """Return whether mapping is needed.

        There are two situations where we must use the mapping:
        - this isn't the default driver (i.e. multiple backends), or
        - we have a single backend that doesn't use UUIDs
        The exception to the above is that we must honor backward
        compatibility if this is the default driver (e.g. to support
        current LDAP)
        """
        is_not_default_driver = driver is not self.driver
        return is_not_default_driver or (
            not driver.generates_uuids()
            and not CONF.identity_mapping.backward_compatible_ids
        )

    def _clear_domain_id_if_domain_unaware(self, driver, ref):
        """Clear domain_id details if driver is not domain aware."""
        if not driver.is_domain_aware() and 'domain_id' in ref:
            ref = ref.copy()
            ref.pop('domain_id')
        return ref

    def _select_identity_driver(self, domain_id):
        """Choose a backend driver for the given domain_id.

        :param domain_id: The domain_id for which we want to find a driver.  If
                          the domain_id is specified as None, then this means
                          we need a driver that handles multiple domains.

        :returns: chosen backend driver

        If there is a specific driver defined for this domain then choose it.
        If the domain is None, or there no specific backend for the given
        domain is found, then we chose the default driver.

        """
        if domain_id is None:
            driver = self.driver
        else:
            driver = (
                self.domain_configs.get_domain_driver(domain_id) or self.driver
            )

        # If the driver is not domain aware (e.g. LDAP) then check to
        # ensure we are not mapping multiple domains onto it - the only way
        # that would happen is that the default driver is LDAP and the
        # domain is anything other than None or the default domain.
        if (
            not driver.is_domain_aware()
            and driver == self.driver
            and domain_id != CONF.identity.default_domain_id
            and domain_id is not None
        ):
            LOG.warning(
                'Found multiple domains being mapped to a '
                'driver that does not support that (e.g. '
                'LDAP) - Domain ID: %(domain)s, '
                'Default Driver: %(driver)s',
                {'domain': domain_id, 'driver': (driver == self.driver)},
            )
            raise exception.DomainNotFound(domain_id=domain_id)
        return driver

    def _get_domain_driver_and_entity_id(self, public_id):
        """Look up details using the public ID.

        :param public_id: the ID provided in the call

        :returns: domain_id, which can be None to indicate that the driver
                  in question supports multiple domains
                  driver selected based on this domain
                  entity_id which will is understood by the driver.

        Use the mapping table to look up the domain, driver and local entity
        that is represented by the provided public ID.  Handle the situations
        where we do not use the mapping (e.g. single driver that understands
        UUIDs etc.)

        """
        conf = CONF.identity
        # First, since we don't know anything about the entity yet, we must
        # assume it needs mapping, so long as we are using domain specific
        # drivers.
        if conf.domain_specific_drivers_enabled:
            local_id_ref = PROVIDERS.id_mapping_api.get_id_mapping(public_id)
            if local_id_ref:
                return (
                    local_id_ref['domain_id'],
                    self._select_identity_driver(local_id_ref['domain_id']),
                    local_id_ref['local_id'],
                )

        # So either we are using multiple drivers but the public ID is invalid
        # (and hence was not found in the mapping table), or the public ID is
        # being handled by the default driver.  Either way, the only place left
        # to look is in that standard driver. However, we don't yet know if
        # this driver also needs mapping (e.g. LDAP in non backward
        # compatibility mode).
        driver = self.driver
        if driver.generates_uuids():
            if driver.is_domain_aware:
                # No mapping required, and the driver can handle the domain
                # information itself.  The classic case of this is the
                # current SQL driver.
                return (None, driver, public_id)
            else:
                # Although we don't have any drivers of this type, i.e. that
                # understand UUIDs but not domains, conceptually you could.
                return (conf.default_domain_id, driver, public_id)

        # So the only place left to find the ID is in the default driver which
        # we now know doesn't generate UUIDs
        if not CONF.identity_mapping.backward_compatible_ids:
            # We are not running in backward compatibility mode, so we
            # must use a mapping.
            local_id_ref = PROVIDERS.id_mapping_api.get_id_mapping(public_id)
            if local_id_ref:
                return (
                    local_id_ref['domain_id'],
                    driver,
                    local_id_ref['local_id'],
                )
            else:
                raise exception.PublicIDNotFound(id=public_id)

        # If we reach here, this means that the default driver
        # requires no mapping - but also doesn't understand domains
        # (e.g. the classic single LDAP driver situation). Hence we pass
        # back the public_ID unmodified and use the default domain (to
        # keep backwards compatibility with existing installations).
        #
        # It is still possible that the public ID is just invalid in
        # which case we leave this to the caller to check.
        return (conf.default_domain_id, driver, public_id)

    def _assert_user_and_group_in_same_backend(
        self, user_entity_id, user_driver, group_entity_id, group_driver
    ):
        """Ensure that user and group IDs are backed by the same backend.

        Raise a CrossBackendNotAllowed exception if they are not from the same
        backend, otherwise return None.

        """
        if user_driver is not group_driver:
            # Determine first if either IDs don't exist by calling
            # the driver.get methods (which will raise a NotFound
            # exception).
            user_driver.get_user(user_entity_id)
            group_driver.get_group(group_entity_id)
            # If we get here, then someone is attempting to create a cross
            # backend membership, which is not allowed.
            raise exception.CrossBackendNotAllowed(
                group_id=group_entity_id, user_id=user_entity_id
            )

    def _mark_domain_id_filter_satisfied(self, hints):
        if hints:
            for filter in hints.filters:
                if (
                    filter['name'] == 'domain_id'
                    and filter['comparator'] == 'equals'
                ):
                    hints.filters.remove(filter)

    def _ensure_domain_id_in_hints(self, hints, domain_id):
        if domain_id is not None and not hints.get_exact_filter_by_name(
            'domain_id'
        ):
            hints.add_filter('domain_id', domain_id)

    def _set_list_limit_in_hints(self, hints, driver):
        """Set list limit in hints from driver.

        If a hints list is provided, the wrapper will insert the relevant
        limit into the hints so that the underlying driver call can try and
        honor it. If the driver does truncate the response, it will update the
        'truncated' attribute in the 'limit' entry in the hints list, which
        enables the caller of this function to know if truncation has taken
        place. If, however, the driver layer is unable to perform truncation,
        the 'limit' entry is simply left in the hints list for the caller to
        handle.

        A _get_list_limit() method is required to be present in the object
        class hierarchy, which returns the limit for this backend to which
        we will truncate.

        If a hints list is not provided in the arguments of the wrapped call
        then any limits set in the config file are ignored.  This allows
        internal use of such wrapped methods where the entire data set is
        needed as input for the calculations of some other API (e.g. get role
        assignments for a given project).

        This method, specific to identity manager, is used instead of more
        general response_truncated, because the limit for identity entities
        can be overridden in domain-specific config files. The driver to use
        is determined during processing of the passed parameters and
        response_truncated is designed to set the limit before any processing.
        """
        if hints is None:
            return

        list_limit = driver._get_list_limit()
        if list_limit:
            hints.set_limit(list_limit)

    # The actual driver calls - these are pre/post processed here as
    # part of the Manager layer to make sure we:
    #
    # - select the right driver for this domain
    # - clear/set domain_ids for drivers that do not support domains
    # - create any ID mapping that might be required
    @notifications.emit_event('authenticate')
    @domains_configured
    @exception_translated('assertion')
    def authenticate(self, user_id, password):
        domain_id, driver, entity_id = self._get_domain_driver_and_entity_id(
            user_id
        )
        ref = driver.authenticate(entity_id, password)
        ref = self._set_domain_id_and_mapping(
            ref, domain_id, driver, mapping.EntityType.USER
        )
        ref = self._shadow_nonlocal_user(ref)
        PROVIDERS.shadow_users_api.set_last_active_at(ref['id'])
        return ref

    def _assert_default_project_id_is_not_domain(self, default_project_id):
        if default_project_id:
            # make sure project is not a domain
            try:
                project_ref = PROVIDERS.resource_api.get_project(
                    default_project_id
                )
                if project_ref['is_domain'] is True:
                    msg = _(
                        "User's default project ID cannot be a "
                        "domain ID: %s"
                    )
                    raise exception.ValidationError(
                        message=(msg % default_project_id)
                    )
            except exception.ProjectNotFound:
                # should be idempotent if project is not found so that it is
                # backward compatible
                pass

    def _validate_federated_objects(self, fed_obj_list):
        # Validate that the ipd and protocols exist
        for fed_obj in fed_obj_list:
            try:
                self.federation_api.get_idp(fed_obj['idp_id'])
            except exception.IdentityProviderNotFound:
                msg = (
                    _("Could not find Identity Provider: %s")
                    % fed_obj['idp_id']
                )
                raise exception.ValidationError(msg)
            for protocol in fed_obj['protocols']:
                try:
                    self.federation_api.get_protocol(
                        fed_obj['idp_id'], protocol['protocol_id']
                    )
                except exception.FederatedProtocolNotFound:
                    msg = _(
                        "Could not find federated protocol "
                        "%(protocol)s for Identity Provider: %(idp)s."
                    ) % {
                        'protocol': protocol['protocol_id'],
                        'idp': fed_obj['idp_id'],
                    }
                    raise exception.ValidationError(msg)

    def _create_federated_objects(self, user_ref, fed_obj_list):
        for fed_obj in fed_obj_list:
            for protocols in fed_obj['protocols']:
                federated_dict = {
                    'user_id': user_ref['id'],
                    'idp_id': fed_obj['idp_id'],
                    'protocol_id': protocols['protocol_id'],
                    'unique_id': protocols['unique_id'],
                    'display_name': user_ref['name'],
                }
                self.shadow_users_api.create_federated_object(federated_dict)

    def _create_user_with_federated_objects(self, user, driver):
        # If the user did not pass a federated object along inside the user
        # object then we simply create the user as normal.
        if not user.get('federated'):
            if 'federated' in user:
                del user['federated']
            user = driver.create_user(user['id'], user)
            return user
        # Otherwise, validate the federated object and create the user.
        else:
            user_ref = user.copy()
            del user['federated']
            self._validate_federated_objects(user_ref['federated'])
            user = driver.create_user(user['id'], user)
            self._create_federated_objects(user_ref, user_ref['federated'])
            user['federated'] = user_ref['federated']
            return user

    @domains_configured
    @exception_translated('user')
    def create_user(self, user_ref, initiator=None):
        user = user_ref.copy()
        if 'password' in user:
            validators.validate_password(user['password'])
        user['name'] = user['name'].strip()
        user.setdefault('enabled', True)
        domain_id = user['domain_id']
        PROVIDERS.resource_api.get_domain(domain_id)

        self._assert_default_project_id_is_not_domain(
            user_ref.get('default_project_id')
        )

        # For creating a user, the domain is in the object itself
        domain_id = user_ref['domain_id']
        driver = self._select_identity_driver(domain_id)
        user = self._clear_domain_id_if_domain_unaware(driver, user)
        # Generate a local ID - in the future this might become a function of
        # the underlying driver so that it could conform to rules set down by
        # that particular driver type.
        user['id'] = uuid.uuid4().hex
        ref = self._create_user_with_federated_objects(user, driver)
        notifications.Audit.created(self._USER, user['id'], initiator)
        return self._set_domain_id_and_mapping(
            ref, domain_id, driver, mapping.EntityType.USER
        )

    @domains_configured
    @exception_translated('user')
    @MEMOIZE
    def get_user(self, user_id):
        domain_id, driver, entity_id = self._get_domain_driver_and_entity_id(
            user_id
        )
        ref = driver.get_user(entity_id)
        # Add user's federated objects
        fed_objects = self.shadow_users_api.get_federated_objects(user_id)
        if fed_objects:
            ref['federated'] = fed_objects
        return self._set_domain_id_and_mapping(
            ref, domain_id, driver, mapping.EntityType.USER
        )

    def assert_user_enabled(self, user_id, user=None):
        """Assert the user and the user's domain are enabled.

        :raise AssertionError if the user or the user's domain is disabled.
        """
        if user is None:
            user = self.get_user(user_id)
        PROVIDERS.resource_api.assert_domain_enabled(user['domain_id'])
        if not user.get('enabled', True):
            raise AssertionError(_('User is disabled: %s') % user_id)

    @domains_configured
    @exception_translated('user')
    @MEMOIZE
    def get_user_by_name(self, user_name, domain_id):
        driver = self._select_identity_driver(domain_id)
        ref = driver.get_user_by_name(user_name, domain_id)
        return self._set_domain_id_and_mapping(
            ref, domain_id, driver, mapping.EntityType.USER
        )

    def _translate_expired_password_hints(self, hints):
        """Clean Up Expired Password Hints.

        Any `password_expires_at` filters on the `list_users` or
        `list_users_in_group` queries are modified so the call will
        return valid data.

        The filters `comparator` is changed to the operator specified in
        the call, otherwise it is assumed to be `equals`. The filters
        `value` becomes the timestamp specified. Both the operator and
        timestamp are validated, and will raise a InvalidOperatorError
        or ValidationTimeStampError exception respectively if invalid.

        """
        operators = {
            'lt': operator.lt,
            'gt': operator.gt,
            'eq': operator.eq,
            'lte': operator.le,
            'gte': operator.ge,
            'neq': operator.ne,
        }
        for filter_ in hints.filters:
            if 'password_expires_at' == filter_['name']:
                # password_expires_at must be in the format
                # 'lt:2016-11-06T15:32:17Z'. So we can assume the position
                # of the ':' otherwise assign the operator to equals.
                if ':' in filter_['value'][2:4]:
                    op, timestamp = filter_['value'].split(':', 1)
                else:
                    op = 'eq'
                    timestamp = filter_['value']

                try:
                    filter_['value'] = timeutils.parse_isotime(timestamp)
                except ValueError:
                    raise exception.ValidationTimeStampError

                try:
                    filter_['comparator'] = operators[op]
                except KeyError:
                    raise exception.InvalidOperatorError(_op=op)
        return hints

    def _handle_shadow_and_local_users(self, driver, hints):
        federated_attributes = {'idp_id', 'protocol_id', 'unique_id'}
        fed_res = []
        for filter_ in hints.filters:
            if filter_['name'] in federated_attributes:
                return PROVIDERS.shadow_users_api.get_federated_users(hints)
            # Note: If the filters contain 'name', we should get the user from
            # both local user and shadow user backend.
            if filter_['name'] == 'name':
                fed_hints = copy.deepcopy(hints)
                fed_res = PROVIDERS.shadow_users_api.get_federated_users(
                    fed_hints
                )
                break
        return driver.list_users(hints) + fed_res

    @domains_configured
    @exception_translated('user')
    def list_users(self, domain_scope=None, hints=None):
        driver = self._select_identity_driver(domain_scope)
        self._set_list_limit_in_hints(hints, driver)
        hints = hints or driver_hints.Hints()
        if driver.is_domain_aware():
            # Force the domain_scope into the hint to ensure that we only get
            # back domains for that scope.
            self._ensure_domain_id_in_hints(hints, domain_scope)
        else:
            # We are effectively satisfying any domain_id filter by the above
            # driver selection, so remove any such filter.
            self._mark_domain_id_filter_satisfied(hints)
        hints = self._translate_expired_password_hints(hints)
        ref_list = self._handle_shadow_and_local_users(driver, hints)
        return self._set_domain_id_and_mapping(
            ref_list, domain_scope, driver, mapping.EntityType.USER
        )

    def _require_matching_domain_id(self, new_ref, orig_ref):
        """Ensure the current domain ID matches the reference one, if any.

        Provided we want domain IDs to be immutable, check whether any
        domain_id specified in the ref dictionary matches the existing
        domain_id for this entity.

        :param new_ref: the dictionary of new values proposed for this entity
        :param orig_ref: the dictionary of original values proposed for this
                         entity
        :raises: :class:`keystone.exception.ValidationError`
        """
        if 'domain_id' in new_ref:
            if new_ref['domain_id'] != orig_ref['domain_id']:
                raise exception.ValidationError(_('Cannot change Domain ID'))

    def _update_user_with_federated_objects(self, user, driver, entity_id):
        # If the user did not pass a federated object along inside the user
        # object then we simply update the user as normal and add the
        # currently associated federated objects to user to be added to the
        # dictionary.
        if not user.get('federated'):
            if 'federated' in user:
                del user['federated']
            user = driver.update_user(entity_id, user)
            fed_objects = self.shadow_users_api.get_federated_objects(
                user['id']
            )
            if fed_objects:
                user['federated'] = fed_objects
            return user
        # Otherwise, we validate, remove the previous user's federated objects,
        # and update the user along with their updated federated objects.
        else:
            user_ref = user.copy()
            self._validate_federated_objects(user_ref['federated'])
            self.shadow_users_api.delete_federated_object(entity_id)
            del user['federated']
            user = driver.update_user(entity_id, user)
            self._create_federated_objects(user, user_ref['federated'])
            user['federated'] = user_ref['federated']
            return user

    @domains_configured
    @exception_translated('user')
    def update_user(self, user_id, user_ref, initiator=None):
        old_user_ref = self.get_user(user_id)
        user = user_ref.copy()
        self._require_matching_domain_id(user, old_user_ref)
        if 'password' in user:
            validators.validate_password(user['password'])
        if 'name' in user:
            user['name'] = user['name'].strip()
        if 'id' in user:
            if user_id != user['id']:
                raise exception.ValidationError(_('Cannot change user ID'))
            # Since any ID in the user dict is now irrelevant, remove its so as
            # the driver layer won't be confused by the fact the this is the
            # public ID not the local ID
            user.pop('id')

        self._assert_default_project_id_is_not_domain(
            user_ref.get('default_project_id')
        )

        domain_id, driver, entity_id = self._get_domain_driver_and_entity_id(
            user_id
        )
        user = self._clear_domain_id_if_domain_unaware(driver, user)
        self.get_user.invalidate(self, old_user_ref['id'])
        self.get_user_by_name.invalidate(
            self, old_user_ref['name'], old_user_ref['domain_id']
        )

        ref = self._update_user_with_federated_objects(user, driver, entity_id)

        notifications.Audit.updated(self._USER, user_id, initiator)

        enabled_change = (user.get('enabled') is False) and user[
            'enabled'
        ] != old_user_ref.get('enabled')
        if enabled_change or user.get('password') is not None:
            self._persist_revocation_event_for_user(user_id)
            reason = (
                f'Invalidating the token cache because user {user_id} was '
                'enabled or disabled. Authorization will be calculated and '
                'enforced accordingly the next time they authenticate or '
                'validate a token.'
            )
            notifications.invalidate_token_cache_notification(reason)

        return self._set_domain_id_and_mapping(
            ref, domain_id, driver, mapping.EntityType.USER
        )

    @domains_configured
    @exception_translated('user')
    def delete_user(self, user_id, initiator=None):
        domain_id, driver, entity_id = self._get_domain_driver_and_entity_id(
            user_id
        )
        # Get user details to invalidate the cache.
        user_old = self.get_user(user_id)

        hints = driver_hints.Hints()
        hints.add_filter('user_id', user_id)

        driver.delete_user(entity_id)
        PROVIDERS.assignment_api.delete_user_assignments(user_id)
        self.get_user.invalidate(self, user_id)
        self.get_user_by_name.invalidate(
            self, user_old['name'], user_old['domain_id']
        )

        PROVIDERS.credential_api.delete_credentials_for_user(user_id)
        PROVIDERS.id_mapping_api.delete_id_mapping(user_id)
        notifications.Audit.deleted(self._USER, user_id, initiator)

        # Invalidate user role assignments cache region, as it may be caching
        # role assignments where the actor is the specified user
        assignment.COMPUTED_ASSIGNMENTS_REGION.invalidate()

    @domains_configured
    @exception_translated('group')
    def create_group(self, group_ref, initiator=None):
        group = group_ref.copy()
        group.setdefault('description', '')
        domain_id = group['domain_id']
        PROVIDERS.resource_api.get_domain(domain_id)

        # For creating a group, the domain is in the object itself
        domain_id = group_ref['domain_id']
        driver = self._select_identity_driver(domain_id)
        group = self._clear_domain_id_if_domain_unaware(driver, group)
        # Generate a local ID - in the future this might become a function of
        # the underlying driver so that it could conform to rules set down by
        # that particular driver type.
        group['id'] = uuid.uuid4().hex
        group['name'] = group['name'].strip()
        ref = driver.create_group(group['id'], group)

        notifications.Audit.created(self._GROUP, group['id'], initiator)

        return self._set_domain_id_and_mapping(
            ref, domain_id, driver, mapping.EntityType.GROUP
        )

    @domains_configured
    @exception_translated('group')
    @MEMOIZE
    def get_group(self, group_id):
        domain_id, driver, entity_id = self._get_domain_driver_and_entity_id(
            group_id
        )
        ref = driver.get_group(entity_id)
        return self._set_domain_id_and_mapping(
            ref, domain_id, driver, mapping.EntityType.GROUP
        )

    @domains_configured
    @exception_translated('group')
    def get_group_by_name(self, group_name, domain_id):
        driver = self._select_identity_driver(domain_id)
        ref = driver.get_group_by_name(group_name, domain_id)
        return self._set_domain_id_and_mapping(
            ref, domain_id, driver, mapping.EntityType.GROUP
        )

    @domains_configured
    @exception_translated('group')
    def update_group(self, group_id, group, initiator=None):
        old_group_ref = self.get_group(group_id)
        self._require_matching_domain_id(group, old_group_ref)
        domain_id, driver, entity_id = self._get_domain_driver_and_entity_id(
            group_id
        )
        group = self._clear_domain_id_if_domain_unaware(driver, group)
        if 'name' in group:
            group['name'] = group['name'].strip()
        ref = driver.update_group(entity_id, group)
        self.get_group.invalidate(self, group_id)
        notifications.Audit.updated(self._GROUP, group_id, initiator)
        return self._set_domain_id_and_mapping(
            ref, domain_id, driver, mapping.EntityType.GROUP
        )

    @domains_configured
    @exception_translated('group')
    def delete_group(self, group_id, initiator=None):
        domain_id, driver, entity_id = self._get_domain_driver_and_entity_id(
            group_id
        )
        roles = PROVIDERS.assignment_api.list_role_assignments(
            group_id=group_id
        )
        user_ids = (u['id'] for u in self.list_users_in_group(group_id))
        driver.delete_group(entity_id)
        self.get_group.invalidate(self, group_id)
        PROVIDERS.id_mapping_api.delete_id_mapping(group_id)
        PROVIDERS.assignment_api.delete_group_assignments(group_id)

        notifications.Audit.deleted(self._GROUP, group_id, initiator)

        # If the group has been created and has users but has no role
        # assignment for the group then we do not need to revoke all the users
        # tokens and can just delete the group.
        if roles:
            for user_id in user_ids:
                self._persist_revocation_event_for_user(user_id)

            reason_s = (
                f'Invalidating the token cache because group {group_id} '
                'has been deleted.'
            )
            notifications.invalidate_token_cache_notification(reason_s)

        # Invalidate user role assignments cache region, as it may be caching
        # role assignments expanded from the specified group to its users
        assignment.COMPUTED_ASSIGNMENTS_REGION.invalidate()

    @domains_configured
    @exception_translated('group')
    def add_user_to_group(self, user_id, group_id, initiator=None):
        @exception_translated('user')
        def get_entity_info_for_user(public_id):
            return self._get_domain_driver_and_entity_id(public_id)

        _domain_id, group_driver, group_entity_id = (
            self._get_domain_driver_and_entity_id(group_id)
        )
        # Get the same info for the user_id, taking care to map any
        # exceptions correctly
        _domain_id, user_driver, user_entity_id = get_entity_info_for_user(
            user_id
        )

        self._assert_user_and_group_in_same_backend(
            user_entity_id, user_driver, group_entity_id, group_driver
        )

        group_driver.add_user_to_group(user_entity_id, group_entity_id)

        # Invalidate user role assignments cache region, as it may now need to
        # include role assignments from the specified group to its users
        assignment.COMPUTED_ASSIGNMENTS_REGION.invalidate()
        notifications.Audit.added_to(
            self._GROUP, group_id, self._USER, user_id, initiator
        )

    @domains_configured
    @exception_translated('group')
    def remove_user_from_group(self, user_id, group_id, initiator=None):
        @exception_translated('user')
        def get_entity_info_for_user(public_id):
            return self._get_domain_driver_and_entity_id(public_id)

        _domain_id, group_driver, group_entity_id = (
            self._get_domain_driver_and_entity_id(group_id)
        )
        # Get the same info for the user_id, taking care to map any
        # exceptions correctly
        _domain_id, user_driver, user_entity_id = get_entity_info_for_user(
            user_id
        )

        self._assert_user_and_group_in_same_backend(
            user_entity_id, user_driver, group_entity_id, group_driver
        )

        group_driver.remove_user_from_group(user_entity_id, group_entity_id)
        self._persist_revocation_event_for_user(user_id)

        # Invalidate user role assignments cache region, as it may be caching
        # role assignments expanded from this group to this user
        assignment.COMPUTED_ASSIGNMENTS_REGION.invalidate()
        reason = (
            f'Invalidating the token cache because user {user_id} was '
            f'removed from group {group_id}. Authorization will be '
            'calculated and enforced accordingly the next time they '
            'authenticate or validate a token.'
        )
        notifications.invalidate_token_cache_notification(reason)
        notifications.Audit.removed_from(
            self._GROUP, group_id, self._USER, user_id, initiator
        )

    def _persist_revocation_event_for_user(self, user_id):
        """Emit a notification to invoke a revocation event callback.

        Fire off an internal notification that will be consumed by the
        revocation API to store a revocation record for a specific user.

        :param user_id: user identifier
        :type user_id: string
        """
        notifications.Audit.internal(
            notifications.PERSIST_REVOCATION_EVENT_FOR_USER, user_id
        )

    @domains_configured
    @exception_translated('user')
    def list_groups_for_user(self, user_id, hints=None):
        domain_id, driver, entity_id = self._get_domain_driver_and_entity_id(
            user_id
        )
        self._set_list_limit_in_hints(hints, driver)
        hints = hints or driver_hints.Hints()
        if not driver.is_domain_aware():
            # We are effectively satisfying any domain_id filter by the above
            # driver selection, so remove any such filter
            self._mark_domain_id_filter_satisfied(hints)
        ref_list = driver.list_groups_for_user(entity_id, hints)
        for ref in ref_list:
            if 'membership_expires_at' not in ref:
                ref['membership_expires_at'] = None
        return self._set_domain_id_and_mapping(
            ref_list, domain_id, driver, mapping.EntityType.GROUP
        )

    @domains_configured
    @exception_translated('group')
    def list_groups(self, domain_scope=None, hints=None):
        driver = self._select_identity_driver(domain_scope)
        self._set_list_limit_in_hints(hints, driver)
        hints = hints or driver_hints.Hints()
        if driver.is_domain_aware():
            # Force the domain_scope into the hint to ensure that we only get
            # back domains for that scope.
            self._ensure_domain_id_in_hints(hints, domain_scope)
        else:
            # We are effectively satisfying any domain_id filter by the above
            # driver selection, so remove any such filter.
            self._mark_domain_id_filter_satisfied(hints)
        ref_list = driver.list_groups(hints)
        return self._set_domain_id_and_mapping(
            ref_list, domain_scope, driver, mapping.EntityType.GROUP
        )

    @domains_configured
    @exception_translated('group')
    def list_users_in_group(self, group_id, hints=None):
        domain_id, driver, entity_id = self._get_domain_driver_and_entity_id(
            group_id
        )
        self._set_list_limit_in_hints(hints, driver)
        hints = hints or driver_hints.Hints()
        if not driver.is_domain_aware():
            # We are effectively satisfying any domain_id filter by the above
            # driver selection, so remove any such filter
            self._mark_domain_id_filter_satisfied(hints)
        hints = self._translate_expired_password_hints(hints)
        ref_list = driver.list_users_in_group(entity_id, hints)
        return self._set_domain_id_and_mapping(
            ref_list, domain_id, driver, mapping.EntityType.USER
        )

    @domains_configured
    @exception_translated('group')
    def check_user_in_group(self, user_id, group_id):
        @exception_translated('user')
        def get_entity_info_for_user(public_id):
            return self._get_domain_driver_and_entity_id(public_id)

        _domain_id, group_driver, group_entity_id = (
            self._get_domain_driver_and_entity_id(group_id)
        )
        # Get the same info for the user_id, taking care to map any
        # exceptions correctly
        _domain_id, user_driver, user_entity_id = get_entity_info_for_user(
            user_id
        )

        self._assert_user_and_group_in_same_backend(
            user_entity_id, user_driver, group_entity_id, group_driver
        )

        return group_driver.check_user_in_group(
            user_entity_id, group_entity_id
        )

    @domains_configured
    def change_password(
        self, user_id, original_password, new_password, initiator=None
    ):
        # authenticate() will raise an AssertionError if authentication fails
        try:
            self.authenticate(user_id, original_password)
        except exception.PasswordExpired:
            # If a password has expired, we want users to be able to change it
            pass

        domain_id, driver, entity_id = self._get_domain_driver_and_entity_id(
            user_id
        )
        try:
            validators.validate_password(new_password)
            driver.change_password(entity_id, new_password)
        except exception.PasswordValidationError as ex:
            audit_reason = reason.Reason(str(ex), str(ex.code))
            notifications.Audit.updated(
                self._USER, user_id, initiator, reason=audit_reason
            )
            raise

        notifications.Audit.updated(self._USER, user_id, initiator)
        self._persist_revocation_event_for_user(user_id)
        reason_s = (
            f'Invalidating the token cache because user {user_id} changed '
            'the password. Authorization will be calculated and enforced '
            'accordingly the next time they authenticate or validate a '
            'token.'
        )
        notifications.invalidate_token_cache_notification(reason_s)

    @MEMOIZE
    def _shadow_nonlocal_user(self, user):
        try:
            return PROVIDERS.shadow_users_api.get_user(user['id'])
        except exception.UserNotFound:
            return PROVIDERS.shadow_users_api.create_nonlocal_user(user)

    def _shadow_federated_user(self, idp_id, protocol_id, user):
        user_dict = {}
        email = user.get('email')
        try:
            LOG.debug("Trying to update name for federated user [%s].", user)

            PROVIDERS.shadow_users_api.update_federated_user_display_name(
                idp_id, protocol_id, user['id'], user['name']
            )
            user_dict = PROVIDERS.shadow_users_api.get_federated_user(
                idp_id, protocol_id, user['id']
            )

            if email:
                LOG.debug(
                    "Executing the e-mail update for federated user [%s].",
                    user,
                )

                user_ref = {"email": email}
                self.update_user(user_dict['id'], user_ref)
                user_dict.update({"email": email})
        except exception.UserNotFound:
            federated_dict = {
                'idp_id': idp_id,
                'protocol_id': protocol_id,
                'unique_id': user['id'],
                'display_name': user['name'],
            }
            LOG.debug("Creating federated user [%s].", user)
            user_dict = PROVIDERS.shadow_users_api.create_federated_user(
                user["domain"]['id'], federated_dict, email=email
            )
        PROVIDERS.shadow_users_api.set_last_active_at(user_dict['id'])
        return user_dict

    def shadow_federated_user(self, idp_id, protocol_id, user, group_ids=None):
        """Map a federated user to a user.

        :param idp_id: identity provider id
        :param protocol_id: protocol id
        :param user: User dictionary
        :param group_ids: list of group ids to add the user to

        :returns: dictionary of the mapped User entity
        """
        user_dict = self._shadow_federated_user(idp_id, protocol_id, user)

        # Note(knikolla): The shadowing operation can be cached,
        # however we need to update the expiring group memberships.
        if group_ids:
            for group_id in group_ids:
                LOG.info(
                    "Adding user [%s] to group [%s].", user_dict, group_id
                )
                PROVIDERS.shadow_users_api.add_user_to_group_expires(
                    user_dict['id'], group_id
                )
        return user_dict


class MappingManager(manager.Manager):
    """Default pivot point for the ID Mapping backend."""

    driver_namespace = 'keystone.identity.id_mapping'
    _provides_api = 'id_mapping_api'

    def __init__(self):
        super().__init__(CONF.identity_mapping.driver)

    @MEMOIZE_ID_MAPPING
    def _get_public_id(self, domain_id, local_id, entity_type):
        return self.driver.get_public_id(
            {
                'domain_id': domain_id,
                'local_id': local_id,
                'entity_type': entity_type,
            }
        )

    def get_public_id(self, local_entity):
        return self._get_public_id(
            local_entity['domain_id'],
            local_entity['local_id'],
            local_entity['entity_type'],
        )

    @MEMOIZE_ID_MAPPING
    def get_id_mapping(self, public_id):
        return self.driver.get_id_mapping(public_id)

    def create_id_mapping(self, local_entity, public_id=None):
        public_id = self.driver.create_id_mapping(local_entity, public_id)
        if MEMOIZE_ID_MAPPING.should_cache(public_id):
            self._get_public_id.set(
                public_id,
                self,
                local_entity['domain_id'],
                local_entity['local_id'],
                local_entity['entity_type'],
            )
            self.get_id_mapping.set(local_entity, self, public_id)
        return public_id

    def delete_id_mapping(self, public_id):
        local_entity = self.get_id_mapping.get(self, public_id)
        self.driver.delete_id_mapping(public_id)
        # Delete the key of entity from cache
        if local_entity:
            self._get_public_id.invalidate(
                self,
                local_entity['domain_id'],
                local_entity['local_id'],
                local_entity['entity_type'],
            )
        self.get_id_mapping.invalidate(self, public_id)

    def purge_mappings(self, purge_filter):
        # Purge mapping is rarely used and only used by the command client,
        # it's quite complex to invalidate part of the cache based on the purge
        # filters, so here invalidate the whole cache when purging mappings.
        self.driver.purge_mappings(purge_filter)
        ID_MAPPING_REGION.invalidate()


class ShadowUsersManager(manager.Manager):
    """Default pivot point for the Shadow Users backend."""

    driver_namespace = 'keystone.identity.shadow_users'
    _provides_api = 'shadow_users_api'

    def __init__(self):
        shadow_driver = CONF.shadow_users.driver

        super().__init__(shadow_driver)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.common import validation
from keystone.common.validation import parameter_types

_consumer_properties = {
    'description': validation.nullable(parameter_types.description)
}

consumer_create = {
    'type': 'object',
    'properties': _consumer_properties,
    'additionalProperties': True,
}

consumer_update = {
    'type': 'object',
    'properties': _consumer_properties,
    'not': {'required': ['secret']},
    'minProperties': 1,
    'additionalProperties': True,
}

request_token_authorize = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'id': parameter_types.id_string,
            'name': parameter_types.name,
        },
        'minProperties': 1,
        'maxProperties': 1,
        'additionalProperties': False,
    },
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Fake driver to test out-of-tree drivers handling."""

from oslo_config import cfg

from keystone import exception
from keystone.identity.backends import base


class FooDriver(base.IdentityDriverBase):
    """Fake out-of-tree driver.

    It does not make much sense to inherit from BaseClass, but in certain
    places across the code methods are invoked

    """

    @classmethod
    def register_opts(cls, conf):
        grp = cfg.OptGroup("foo")
        opts = [cfg.StrOpt("opt1")]
        conf.register_group(grp)
        conf.register_opts(opts, group=grp)

    def authenticate(self, user_id, password):
        raise exception.NotImplemented()  # pragma: no cover

    def create_user(self, user_id, user):
        raise exception.NotImplemented()  # pragma: no cover

    def list_users(self, hints):
        raise exception.NotImplemented()  # pragma: no cover

    def unset_default_project_id(self, project_id):
        raise exception.NotImplemented()  # pragma: no cover

    def list_users_in_group(self, group_id, hints):
        raise exception.NotImplemented()  # pragma: no cover

    def get_user(self, user_id):
        raise exception.NotImplemented()  # pragma: no cover

    def update_user(self, user_id, user):
        raise exception.NotImplemented()  # pragma: no cover

    def change_password(self, user_id, new_password):
        raise exception.NotImplemented()  # pragma: no cover

    def add_user_to_group(self, user_id, group_id):
        raise exception.NotImplemented()  # pragma: no cover

    def check_user_in_group(self, user_id, group_id):
        raise exception.NotImplemented()  # pragma: no cover

    def remove_user_from_group(self, user_id, group_id):
        raise exception.NotImplemented()  # pragma: no cover

    def delete_user(self, user_id):
        raise exception.NotImplemented()  # pragma: no cover

    def get_user_by_name(self, user_name, domain_id):
        raise exception.NotImplemented()  # pragma: no cover

    def reset_last_active(self):
        raise exception.NotImplemented()  # pragma: no cover

    def create_group(self, group_id, group):
        raise exception.NotImplemented()  # pragma: no cover

    def list_groups(self, hints):
        raise exception.NotImplemented()  # pragma: no cover

    def list_groups_for_user(self, user_id, hints):
        raise exception.NotImplemented()  # pragma: no cover

    def get_group(self, group_id):
        raise exception.NotImplemented()  # pragma: no cover

    def get_group_by_name(self, group_name, domain_id):
        raise exception.NotImplemented()  # pragma: no cover

    def update_group(self, group_id, group):
        raise exception.NotImplemented()  # pragma: no cover

    def delete_group(self, group_id):
        raise exception.NotImplemented()  # pragma: no cover



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import datetime
import struct
import uuid

from cryptography import fernet
import msgpack
from oslo_log import log
from oslo_utils import timeutils

from keystone.auth import plugins as auth_plugins
from keystone.common import fernet_utils as utils
from keystone.common import utils as ks_utils
import keystone.conf
from keystone import exception
from keystone.i18n import _

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)

# Fernet byte indexes as computed by pypi/keyless_fernet and defined in
# https://github.com/fernet/spec
TIMESTAMP_START = 1
TIMESTAMP_END = 9


class TokenFormatter:
    """Packs and unpacks payloads into tokens for transport."""

    @property
    def crypto(self):
        """Return a cryptography instance.

        You can extend this class with a custom crypto @property to provide
        your own token encoding / decoding. For example, using a different
        cryptography library (e.g. ``python-keyczar``) or to meet arbitrary
        security requirements.

        This @property just needs to return an object that implements
        ``encrypt(plaintext)`` and ``decrypt(ciphertext)``.

        """
        fernet_utils = utils.FernetUtils(
            CONF.fernet_tokens.key_repository,
            CONF.fernet_tokens.max_active_keys,
            'fernet_tokens',
        )
        keys = fernet_utils.load_keys()

        if not keys:
            raise exception.KeysNotFound()

        fernet_instances = [fernet.Fernet(key) for key in keys]
        return fernet.MultiFernet(fernet_instances)

    def pack(self, payload):
        """Pack a payload for transport as a token.

        :type payload: bytes
        :rtype: str

        """
        # base64 padding (if any) is not URL-safe
        return self.crypto.encrypt(payload).rstrip(b'=').decode('utf-8')

    def unpack(self, token):
        """Unpack a token, and validate the payload.

        :type token: str
        :rtype: bytes

        """
        token = TokenFormatter.restore_padding(token)

        try:
            return self.crypto.decrypt(token.encode('utf-8'))
        except fernet.InvalidToken:
            raise exception.ValidationError(
                _('Could not recognize Fernet token')
            )

    @classmethod
    def restore_padding(cls, token):
        """Restore padding based on token size.

        :param token: token to restore padding on
        :type token: str
        :returns: token with correct padding

        """
        # Re-inflate the padding
        mod_returned = len(token) % 4
        if mod_returned:
            missing_padding = 4 - mod_returned
            token += '=' * missing_padding
        return token

    @classmethod
    def creation_time(cls, fernet_token):
        """Return the creation time of a valid Fernet token.

        :type fernet_token: str

        """
        fernet_token = TokenFormatter.restore_padding(fernet_token)
        # fernet_token is str

        # Fernet tokens are base64 encoded, so we need to unpack them first
        # urlsafe_b64decode() requires bytes
        token_bytes = base64.urlsafe_b64decode(fernet_token.encode('utf-8'))

        # slice into the byte array to get just the timestamp
        timestamp_bytes = token_bytes[TIMESTAMP_START:TIMESTAMP_END]

        # convert those bytes to an integer
        # (it's a 64-bit "unsigned long long int" in C)
        timestamp_int = struct.unpack(">Q", timestamp_bytes)[0]

        # and with an integer, it's trivial to produce a datetime object
        issued_at = datetime.datetime.fromtimestamp(
            timestamp_int, datetime.timezone.utc
        ).replace(tzinfo=None)

        return issued_at

    def create_token(
        self,
        user_id,
        expires_at,
        audit_ids,
        payload_class,
        methods=None,
        system=None,
        domain_id=None,
        project_id=None,
        trust_id=None,
        federated_group_ids=None,
        identity_provider_id=None,
        protocol_id=None,
        access_token_id=None,
        app_cred_id=None,
        thumbprint=None,
    ):
        """Given a set of payload attributes, generate a Fernet token."""
        version = payload_class.version
        payload = payload_class.assemble(
            user_id,
            methods,
            system,
            project_id,
            domain_id,
            expires_at,
            audit_ids,
            trust_id,
            federated_group_ids,
            identity_provider_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
        )

        versioned_payload = (version,) + payload
        serialized_payload = msgpack.packb(versioned_payload)
        token = self.pack(serialized_payload)

        # NOTE(lbragstad): We should warn against Fernet tokens that are over
        # 255 characters in length. This is mostly due to persisting the tokens
        # in a backend store of some kind that might have a limit of 255
        # characters. Even though Keystone isn't storing a Fernet token
        # anywhere, we can't say it isn't being stored somewhere else with
        # those kind of backend constraints.
        if len(token) > CONF.max_token_size:
            LOG.info(
                f'Fernet token created with length of {len(token)} '
                f'characters, which exceeds {CONF.max_token_size} characters'
            )

        return token

    def validate_token(self, token):
        """Validate a Fernet token and returns the payload attributes.

        :type token: str

        """
        serialized_payload = self.unpack(token)
        # TODO(melwitt): msgpack changed their data format in version 1.0, so
        # in order to support a rolling upgrade, we must pass raw=True to
        # support the old format. The try-except may be removed once the
        # N-1 release no longer supports msgpack < 1.0.
        try:
            versioned_payload = msgpack.unpackb(serialized_payload)
        except UnicodeDecodeError:
            versioned_payload = msgpack.unpackb(serialized_payload, raw=True)

        version, payload = versioned_payload[0], versioned_payload[1:]

        for payload_class in _PAYLOAD_CLASSES:
            if version == payload_class.version:
                (
                    user_id,
                    methods,
                    system,
                    project_id,
                    domain_id,
                    expires_at,
                    audit_ids,
                    trust_id,
                    federated_group_ids,
                    identity_provider_id,
                    protocol_id,
                    access_token_id,
                    app_cred_id,
                    thumbprint,
                ) = payload_class.disassemble(payload)
                break
        else:
            # If the token_format is not recognized, raise ValidationError.
            raise exception.ValidationError(
                _('This is not a recognized Fernet payload version: %s')
                % version
            )

        # FIXME(lbragstad): Without this, certain token validation tests fail
        # when running with python 3. Once we get further along in this
        # refactor, we should be better about handling string encoding/types at
        # the edges of the application.
        if isinstance(system, bytes):
            system = system.decode('utf-8')

        # rather than appearing in the payload, the creation time is encoded
        # into the token format itself
        issued_at = TokenFormatter.creation_time(token)
        issued_at = ks_utils.isotime(at=issued_at, subsecond=True)
        expires_at = timeutils.parse_isotime(expires_at)
        expires_at = ks_utils.isotime(at=expires_at, subsecond=True)

        return (
            user_id,
            methods,
            audit_ids,
            system,
            domain_id,
            project_id,
            trust_id,
            federated_group_ids,
            identity_provider_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
            issued_at,
            expires_at,
        )


class BasePayload:
    # each payload variant should have a unique version
    version: int

    @classmethod
    def assemble(
        cls,
        user_id,
        methods,
        system,
        project_id,
        domain_id,
        expires_at,
        audit_ids,
        trust_id,
        federated_group_ids,
        identity_provider_id,
        protocol_id,
        access_token_id,
        app_cred_id,
        thumbprint,
    ):
        """Assemble the payload of a token.

        :param user_id: identifier of the user in the token request
        :param methods: list of authentication methods used
        :param system: a string including system scope information
        :param project_id: ID of the project to scope to
        :param domain_id: ID of the domain to scope to
        :param expires_at: datetime of the token's expiration
        :param audit_ids: list of the token's audit IDs
        :param trust_id: ID of the trust in effect
        :param federated_group_ids: list of group IDs from SAML assertion
        :param identity_provider_id: ID of the user's identity provider
        :param protocol_id: federated protocol used for authentication
        :param access_token_id: ID of the secret in OAuth1 authentication
        :param app_cred_id: ID of the application credential in effect
        :param thumbprint: thumbprint of the certificate in OAuth2 mTLS
        :returns: the payload of a token

        """
        raise NotImplementedError()

    @classmethod
    def disassemble(cls, payload):
        """Disassemble an unscoped payload into the component data.

        The tuple consists of::

            (user_id, methods, system, project_id, domain_id,
             expires_at_str, audit_ids, trust_id, federated_group_ids,
             identity_provider_id, protocol_id,` access_token_id, app_cred_id)

        * ``methods`` are the auth methods.

        Fields will be set to None if they didn't apply to this payload type.

        :param payload: this variant of payload
        :returns: a tuple of the payloads component data

        """
        raise NotImplementedError()

    @classmethod
    def convert_uuid_hex_to_bytes(cls, uuid_string):
        """Compress UUID formatted strings to bytes.

        :param uuid_string: uuid string to compress to bytes
        :returns: a byte representation of the uuid

        """
        uuid_obj = uuid.UUID(uuid_string)
        return uuid_obj.bytes

    @classmethod
    def convert_uuid_bytes_to_hex(cls, uuid_byte_string):
        """Generate uuid.hex format based on byte string.

        :param uuid_byte_string: uuid string to generate from
        :returns: uuid hex formatted string

        """
        uuid_obj = uuid.UUID(bytes=uuid_byte_string)
        return uuid_obj.hex

    @classmethod
    def _convert_time_string_to_float(cls, time_string):
        """Convert a time formatted string to a float.

        :param time_string: time formatted string
        :returns: a timestamp as a float

        """
        time_object = timeutils.parse_isotime(time_string)
        return (
            timeutils.normalize_time(time_object)
            - datetime.datetime.fromtimestamp(
                0, datetime.timezone.utc
            ).replace(tzinfo=None)
        ).total_seconds()

    @classmethod
    def _convert_float_to_time_string(cls, time_float):
        """Convert a floating point timestamp to a string.

        :param time_float: integer representing timestamp
        :returns: a time formatted strings

        """
        time_object = datetime.datetime.fromtimestamp(
            time_float, datetime.timezone.utc
        ).replace(tzinfo=None)
        return ks_utils.isotime(time_object, subsecond=True)

    @classmethod
    def attempt_convert_uuid_hex_to_bytes(cls, value):
        """Attempt to convert value to bytes or return value.

        :param value: value to attempt to convert to bytes
        :returns: tuple containing boolean indicating whether user_id was
                  stored as bytes and uuid value as bytes or the original value

        """
        try:
            return (True, cls.convert_uuid_hex_to_bytes(value))
        except (ValueError, TypeError):
            # ValueError: this might not be a UUID, depending on the
            # situation (i.e. federation)
            # TypeError: the provided value may be binary encoded
            # in which case just return the value (i.e. Python 3)
            return (False, value)

    @classmethod
    def base64_encode(cls, s):
        """Encode a URL-safe string.

        :type s: str
        :rtype: str

        """
        # urlsafe_b64encode() returns bytes so need to convert to
        # str, might as well do it before stripping.
        return base64.urlsafe_b64encode(s).decode('utf-8').rstrip('=')

    @classmethod
    def random_urlsafe_str_to_bytes(cls, s):
        """Convert string from :func:`random_urlsafe_str()` to bytes.

        :type s: str
        :rtype: bytes

        """
        # urlsafe_b64decode() requires str, unicode isn't accepted.
        s = str(s)

        # restore the padding (==) at the end of the string
        return base64.urlsafe_b64decode(s + '==')

    @classmethod
    def _convert_or_decode(cls, is_stored_as_bytes, value):
        """Convert a value to text type, translating uuid -> hex if required.

        :param is_stored_as_bytes: whether value is already bytes
        :type is_stored_as_bytes: boolean
        :param value: value to attempt to convert to bytes
        :type value: str or bytes
        :rtype: str
        """
        if is_stored_as_bytes:
            return cls.convert_uuid_bytes_to_hex(value)
        elif isinstance(value, bytes):
            return value.decode('utf-8')
        return value


class UnscopedPayload(BasePayload):
    version = 0

    @classmethod
    def assemble(
        cls,
        user_id,
        methods,
        system,
        project_id,
        domain_id,
        expires_at,
        audit_ids,
        trust_id,
        federated_group_ids,
        identity_provider_id,
        protocol_id,
        access_token_id,
        app_cred_id,
        thumbprint,
    ):
        b_user_id = cls.attempt_convert_uuid_hex_to_bytes(user_id)
        methods = auth_plugins.convert_method_list_to_integer(methods)
        expires_at_int = cls._convert_time_string_to_float(expires_at)
        b_audit_ids = list(map(cls.random_urlsafe_str_to_bytes, audit_ids))
        return (b_user_id, methods, expires_at_int, b_audit_ids)

    @classmethod
    def disassemble(cls, payload):
        (is_stored_as_bytes, user_id) = payload[0]
        user_id = cls._convert_or_decode(is_stored_as_bytes, user_id)
        methods = auth_plugins.convert_integer_to_method_list(payload[1])
        expires_at_str = cls._convert_float_to_time_string(payload[2])
        audit_ids = list(map(cls.base64_encode, payload[3]))
        system = None
        project_id = None
        domain_id = None
        trust_id = None
        federated_group_ids = None
        identity_provider_id = None
        protocol_id = None
        access_token_id = None
        app_cred_id = None
        thumbprint = None
        return (
            user_id,
            methods,
            system,
            project_id,
            domain_id,
            expires_at_str,
            audit_ids,
            trust_id,
            federated_group_ids,
            identity_provider_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
        )


class DomainScopedPayload(BasePayload):
    version = 1

    @classmethod
    def assemble(
        cls,
        user_id,
        methods,
        system,
        project_id,
        domain_id,
        expires_at,
        audit_ids,
        trust_id,
        federated_group_ids,
        identity_provider_id,
        protocol_id,
        access_token_id,
        app_cred_id,
        thumbprint,
    ):
        b_user_id = cls.attempt_convert_uuid_hex_to_bytes(user_id)
        methods = auth_plugins.convert_method_list_to_integer(methods)
        try:
            b_domain_id = cls.convert_uuid_hex_to_bytes(domain_id)
        except ValueError:
            # the default domain ID is configurable, and probably isn't a UUID
            if domain_id == CONF.identity.default_domain_id:
                b_domain_id = domain_id
            else:
                raise
        expires_at_int = cls._convert_time_string_to_float(expires_at)
        b_audit_ids = list(map(cls.random_urlsafe_str_to_bytes, audit_ids))
        return (b_user_id, methods, b_domain_id, expires_at_int, b_audit_ids)

    @classmethod
    def disassemble(cls, payload):
        (is_stored_as_bytes, user_id) = payload[0]
        user_id = cls._convert_or_decode(is_stored_as_bytes, user_id)
        methods = auth_plugins.convert_integer_to_method_list(payload[1])
        try:
            domain_id = cls.convert_uuid_bytes_to_hex(payload[2])
        except ValueError:
            # the default domain ID is configurable, and probably isn't a UUID
            if isinstance(payload[2], bytes):
                payload[2] = payload[2].decode('utf-8')
            if payload[2] == CONF.identity.default_domain_id:
                domain_id = payload[2]
            else:
                raise
        expires_at_str = cls._convert_float_to_time_string(payload[3])
        audit_ids = list(map(cls.base64_encode, payload[4]))
        system = None
        project_id = None
        trust_id = None
        federated_group_ids = None
        identity_provider_id = None
        protocol_id = None
        access_token_id = None
        app_cred_id = None
        thumbprint = None
        return (
            user_id,
            methods,
            system,
            project_id,
            domain_id,
            expires_at_str,
            audit_ids,
            trust_id,
            federated_group_ids,
            identity_provider_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
        )


class ProjectScopedPayload(BasePayload):
    version = 2

    @classmethod
    def assemble(
        cls,
        user_id,
        methods,
        system,
        project_id,
        domain_id,
        expires_at,
        audit_ids,
        trust_id,
        federated_group_ids,
        identity_provider_id,
        protocol_id,
        access_token_id,
        app_cred_id,
        thumbprint,
    ):
        b_user_id = cls.attempt_convert_uuid_hex_to_bytes(user_id)
        methods = auth_plugins.convert_method_list_to_integer(methods)
        b_project_id = cls.attempt_convert_uuid_hex_to_bytes(project_id)
        expires_at_int = cls._convert_time_string_to_float(expires_at)
        b_audit_ids = list(map(cls.random_urlsafe_str_to_bytes, audit_ids))
        return (b_user_id, methods, b_project_id, expires_at_int, b_audit_ids)

    @classmethod
    def disassemble(cls, payload):
        (is_stored_as_bytes, user_id) = payload[0]
        user_id = cls._convert_or_decode(is_stored_as_bytes, user_id)
        methods = auth_plugins.convert_integer_to_method_list(payload[1])
        (is_stored_as_bytes, project_id) = payload[2]
        project_id = cls._convert_or_decode(is_stored_as_bytes, project_id)
        expires_at_str = cls._convert_float_to_time_string(payload[3])
        audit_ids = list(map(cls.base64_encode, payload[4]))
        system = None
        domain_id = None
        trust_id = None
        federated_group_ids = None
        identity_provider_id = None
        protocol_id = None
        access_token_id = None
        app_cred_id = None
        thumbprint = None
        return (
            user_id,
            methods,
            system,
            project_id,
            domain_id,
            expires_at_str,
            audit_ids,
            trust_id,
            federated_group_ids,
            identity_provider_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
        )


class TrustScopedPayload(BasePayload):
    version = 3

    @classmethod
    def assemble(
        cls,
        user_id,
        methods,
        system,
        project_id,
        domain_id,
        expires_at,
        audit_ids,
        trust_id,
        federated_group_ids,
        identity_provider_id,
        protocol_id,
        access_token_id,
        app_cred_id,
        thumbprint,
    ):
        b_user_id = cls.attempt_convert_uuid_hex_to_bytes(user_id)
        methods = auth_plugins.convert_method_list_to_integer(methods)
        b_project_id = cls.attempt_convert_uuid_hex_to_bytes(project_id)
        b_trust_id = cls.convert_uuid_hex_to_bytes(trust_id)
        expires_at_int = cls._convert_time_string_to_float(expires_at)
        b_audit_ids = list(map(cls.random_urlsafe_str_to_bytes, audit_ids))

        return (
            b_user_id,
            methods,
            b_project_id,
            expires_at_int,
            b_audit_ids,
            b_trust_id,
        )

    @classmethod
    def disassemble(cls, payload):
        (is_stored_as_bytes, user_id) = payload[0]
        user_id = cls._convert_or_decode(is_stored_as_bytes, user_id)
        methods = auth_plugins.convert_integer_to_method_list(payload[1])
        (is_stored_as_bytes, project_id) = payload[2]
        project_id = cls._convert_or_decode(is_stored_as_bytes, project_id)
        expires_at_str = cls._convert_float_to_time_string(payload[3])
        audit_ids = list(map(cls.base64_encode, payload[4]))
        trust_id = cls.convert_uuid_bytes_to_hex(payload[5])
        system = None
        domain_id = None
        federated_group_ids = None
        identity_provider_id = None
        protocol_id = None
        access_token_id = None
        app_cred_id = None
        thumbprint = None
        return (
            user_id,
            methods,
            system,
            project_id,
            domain_id,
            expires_at_str,
            audit_ids,
            trust_id,
            federated_group_ids,
            identity_provider_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
        )


class FederatedUnscopedPayload(BasePayload):
    version = 4

    @classmethod
    def pack_group_id(cls, group_dict):
        return cls.attempt_convert_uuid_hex_to_bytes(group_dict['id'])

    @classmethod
    def unpack_group_id(cls, group_id_in_bytes):
        (is_stored_as_bytes, group_id) = group_id_in_bytes
        group_id = cls._convert_or_decode(is_stored_as_bytes, group_id)
        return {'id': group_id}

    @classmethod
    def assemble(
        cls,
        user_id,
        methods,
        system,
        project_id,
        domain_id,
        expires_at,
        audit_ids,
        trust_id,
        federated_group_ids,
        identity_provider_id,
        protocol_id,
        access_token_id,
        app_cred_id,
        thumbprint,
    ):
        b_user_id = cls.attempt_convert_uuid_hex_to_bytes(user_id)
        methods = auth_plugins.convert_method_list_to_integer(methods)
        b_group_ids = list(map(cls.pack_group_id, federated_group_ids))
        b_idp_id = cls.attempt_convert_uuid_hex_to_bytes(identity_provider_id)
        expires_at_int = cls._convert_time_string_to_float(expires_at)
        b_audit_ids = list(map(cls.random_urlsafe_str_to_bytes, audit_ids))

        return (
            b_user_id,
            methods,
            b_group_ids,
            b_idp_id,
            protocol_id,
            expires_at_int,
            b_audit_ids,
        )

    @classmethod
    def disassemble(cls, payload):
        (is_stored_as_bytes, user_id) = payload[0]
        user_id = cls._convert_or_decode(is_stored_as_bytes, user_id)
        methods = auth_plugins.convert_integer_to_method_list(payload[1])
        group_ids = list(map(cls.unpack_group_id, payload[2]))
        (is_stored_as_bytes, idp_id) = payload[3]
        idp_id = cls._convert_or_decode(is_stored_as_bytes, idp_id)
        protocol_id = payload[4]
        if isinstance(protocol_id, bytes):
            protocol_id = protocol_id.decode('utf-8')
        expires_at_str = cls._convert_float_to_time_string(payload[5])
        audit_ids = list(map(cls.base64_encode, payload[6]))
        system = None
        project_id = None
        domain_id = None
        trust_id = None
        access_token_id = None
        app_cred_id = None
        thumbprint = None
        return (
            user_id,
            methods,
            system,
            project_id,
            domain_id,
            expires_at_str,
            audit_ids,
            trust_id,
            group_ids,
            idp_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
        )


class FederatedScopedPayload(FederatedUnscopedPayload):
    @classmethod
    def assemble(
        cls,
        user_id,
        methods,
        system,
        project_id,
        domain_id,
        expires_at,
        audit_ids,
        trust_id,
        federated_group_ids,
        identity_provider_id,
        protocol_id,
        access_token_id,
        app_cred_id,
        thumbprint,
    ):
        b_user_id = cls.attempt_convert_uuid_hex_to_bytes(user_id)
        methods = auth_plugins.convert_method_list_to_integer(methods)
        b_scope_id = cls.attempt_convert_uuid_hex_to_bytes(
            project_id or domain_id
        )
        b_group_ids = list(map(cls.pack_group_id, federated_group_ids))
        b_idp_id = cls.attempt_convert_uuid_hex_to_bytes(identity_provider_id)
        expires_at_int = cls._convert_time_string_to_float(expires_at)
        b_audit_ids = list(map(cls.random_urlsafe_str_to_bytes, audit_ids))

        return (
            b_user_id,
            methods,
            b_scope_id,
            b_group_ids,
            b_idp_id,
            protocol_id,
            expires_at_int,
            b_audit_ids,
        )

    @classmethod
    def disassemble(cls, payload):
        (is_stored_as_bytes, user_id) = payload[0]
        user_id = cls._convert_or_decode(is_stored_as_bytes, user_id)
        methods = auth_plugins.convert_integer_to_method_list(payload[1])
        (is_stored_as_bytes, scope_id) = payload[2]
        scope_id = cls._convert_or_decode(is_stored_as_bytes, scope_id)
        project_id = (
            scope_id
            if cls.version == FederatedProjectScopedPayload.version
            else None
        )
        domain_id = (
            scope_id
            if cls.version == FederatedDomainScopedPayload.version
            else None
        )
        group_ids = list(map(cls.unpack_group_id, payload[3]))
        (is_stored_as_bytes, idp_id) = payload[4]
        idp_id = cls._convert_or_decode(is_stored_as_bytes, idp_id)
        protocol_id = payload[5]
        if isinstance(protocol_id, bytes):
            protocol_id = protocol_id.decode('utf-8')
        expires_at_str = cls._convert_float_to_time_string(payload[6])
        audit_ids = list(map(cls.base64_encode, payload[7]))
        system = None
        trust_id = None
        access_token_id = None
        app_cred_id = None
        thumbprint = None
        return (
            user_id,
            methods,
            system,
            project_id,
            domain_id,
            expires_at_str,
            audit_ids,
            trust_id,
            group_ids,
            idp_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
        )


class FederatedProjectScopedPayload(FederatedScopedPayload):
    version = 5


class FederatedDomainScopedPayload(FederatedScopedPayload):
    version = 6


class OauthScopedPayload(BasePayload):
    version = 7

    @classmethod
    def assemble(
        cls,
        user_id,
        methods,
        system,
        project_id,
        domain_id,
        expires_at,
        audit_ids,
        trust_id,
        federated_group_ids,
        identity_provider_id,
        protocol_id,
        access_token_id,
        app_cred_id,
        thumbprint,
    ):
        b_user_id = cls.attempt_convert_uuid_hex_to_bytes(user_id)
        methods = auth_plugins.convert_method_list_to_integer(methods)
        b_project_id = cls.attempt_convert_uuid_hex_to_bytes(project_id)
        expires_at_int = cls._convert_time_string_to_float(expires_at)
        b_audit_ids = list(map(cls.random_urlsafe_str_to_bytes, audit_ids))
        b_access_token_id = cls.attempt_convert_uuid_hex_to_bytes(
            access_token_id
        )
        return (
            b_user_id,
            methods,
            b_project_id,
            b_access_token_id,
            expires_at_int,
            b_audit_ids,
        )

    @classmethod
    def disassemble(cls, payload):
        (is_stored_as_bytes, user_id) = payload[0]
        user_id = cls._convert_or_decode(is_stored_as_bytes, user_id)
        methods = auth_plugins.convert_integer_to_method_list(payload[1])
        (is_stored_as_bytes, project_id) = payload[2]
        project_id = cls._convert_or_decode(is_stored_as_bytes, project_id)
        (is_stored_as_bytes, access_token_id) = payload[3]
        access_token_id = cls._convert_or_decode(
            is_stored_as_bytes, access_token_id
        )
        expires_at_str = cls._convert_float_to_time_string(payload[4])
        audit_ids = list(map(cls.base64_encode, payload[5]))
        system = None
        domain_id = None
        trust_id = None
        federated_group_ids = None
        identity_provider_id = None
        protocol_id = None
        app_cred_id = None
        thumbprint = None

        return (
            user_id,
            methods,
            system,
            project_id,
            domain_id,
            expires_at_str,
            audit_ids,
            trust_id,
            federated_group_ids,
            identity_provider_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
        )


class SystemScopedPayload(BasePayload):
    version = 8

    @classmethod
    def assemble(
        cls,
        user_id,
        methods,
        system,
        project_id,
        domain_id,
        expires_at,
        audit_ids,
        trust_id,
        federated_group_ids,
        identity_provider_id,
        protocol_id,
        access_token_id,
        app_cred_id,
        thumbprint,
    ):
        b_user_id = cls.attempt_convert_uuid_hex_to_bytes(user_id)
        methods = auth_plugins.convert_method_list_to_integer(methods)
        expires_at_int = cls._convert_time_string_to_float(expires_at)
        b_audit_ids = list(map(cls.random_urlsafe_str_to_bytes, audit_ids))
        return (b_user_id, methods, system, expires_at_int, b_audit_ids)

    @classmethod
    def disassemble(cls, payload):
        (is_stored_as_bytes, user_id) = payload[0]
        user_id = cls._convert_or_decode(is_stored_as_bytes, user_id)
        methods = auth_plugins.convert_integer_to_method_list(payload[1])
        system = payload[2]
        expires_at_str = cls._convert_float_to_time_string(payload[3])
        audit_ids = list(map(cls.base64_encode, payload[4]))
        project_id = None
        domain_id = None
        trust_id = None
        federated_group_ids = None
        identity_provider_id = None
        protocol_id = None
        access_token_id = None
        app_cred_id = None
        thumbprint = None
        return (
            user_id,
            methods,
            system,
            project_id,
            domain_id,
            expires_at_str,
            audit_ids,
            trust_id,
            federated_group_ids,
            identity_provider_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
        )


class ApplicationCredentialScopedPayload(BasePayload):
    version = 9

    @classmethod
    def assemble(
        cls,
        user_id,
        methods,
        system,
        project_id,
        domain_id,
        expires_at,
        audit_ids,
        trust_id,
        federated_group_ids,
        identity_provider_id,
        protocol_id,
        access_token_id,
        app_cred_id,
        thumbprint,
    ):
        b_user_id = cls.attempt_convert_uuid_hex_to_bytes(user_id)
        methods = auth_plugins.convert_method_list_to_integer(methods)
        b_project_id = cls.attempt_convert_uuid_hex_to_bytes(project_id)
        expires_at_int = cls._convert_time_string_to_float(expires_at)
        b_audit_ids = list(map(cls.random_urlsafe_str_to_bytes, audit_ids))
        b_app_cred_id = cls.attempt_convert_uuid_hex_to_bytes(app_cred_id)
        return (
            b_user_id,
            methods,
            b_project_id,
            expires_at_int,
            b_audit_ids,
            b_app_cred_id,
        )

    @classmethod
    def disassemble(cls, payload):
        (is_stored_as_bytes, user_id) = payload[0]
        user_id = cls._convert_or_decode(is_stored_as_bytes, user_id)
        methods = auth_plugins.convert_integer_to_method_list(payload[1])
        (is_stored_as_bytes, project_id) = payload[2]
        project_id = cls._convert_or_decode(is_stored_as_bytes, project_id)
        expires_at_str = cls._convert_float_to_time_string(payload[3])
        audit_ids = list(map(cls.base64_encode, payload[4]))
        system = None
        domain_id = None
        trust_id = None
        federated_group_ids = None
        identity_provider_id = None
        protocol_id = None
        access_token_id = None
        (is_stored_as_bytes, app_cred_id) = payload[5]
        app_cred_id = cls._convert_or_decode(is_stored_as_bytes, app_cred_id)
        thumbprint = None
        return (
            user_id,
            methods,
            system,
            project_id,
            domain_id,
            expires_at_str,
            audit_ids,
            trust_id,
            federated_group_ids,
            identity_provider_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
        )


class Oauth2CredentialsScopedPayload(BasePayload):
    version = 10

    @classmethod
    def assemble(
        cls,
        user_id,
        methods,
        system,
        project_id,
        domain_id,
        expires_at,
        audit_ids,
        trust_id,
        federated_group_ids,
        identity_provider_id,
        protocol_id,
        access_token_id,
        app_cred_id,
        thumbprint,
    ):
        b_user_id = cls.attempt_convert_uuid_hex_to_bytes(user_id)
        methods = auth_plugins.convert_method_list_to_integer(methods)
        b_project_id = cls.attempt_convert_uuid_hex_to_bytes(project_id)
        b_domain_id = cls.attempt_convert_uuid_hex_to_bytes(domain_id)
        expires_at_int = cls._convert_time_string_to_float(expires_at)
        b_audit_ids = list(map(cls.random_urlsafe_str_to_bytes, audit_ids))
        b_thumbprint = (False, thumbprint)
        return (
            b_user_id,
            methods,
            b_project_id,
            b_domain_id,
            expires_at_int,
            b_audit_ids,
            b_thumbprint,
        )

    @classmethod
    def disassemble(cls, payload):
        (is_stored_as_bytes, user_id) = payload[0]
        user_id = cls._convert_or_decode(is_stored_as_bytes, user_id)
        methods = auth_plugins.convert_integer_to_method_list(payload[1])
        (is_stored_as_bytes, project_id) = payload[2]
        project_id = cls._convert_or_decode(is_stored_as_bytes, project_id)
        (is_stored_as_bytes, domain_id) = payload[3]
        domain_id = cls._convert_or_decode(is_stored_as_bytes, domain_id)
        expires_at_str = cls._convert_float_to_time_string(payload[4])
        audit_ids = list(map(cls.base64_encode, payload[5]))
        (is_stored_as_bytes, thumbprint) = payload[6]
        thumbprint = cls._convert_or_decode(is_stored_as_bytes, thumbprint)
        system = None
        trust_id = None
        federated_group_ids = None
        identity_provider_id = None
        protocol_id = None
        access_token_id = None
        app_cred_id = None
        return (
            user_id,
            methods,
            system,
            project_id,
            domain_id,
            expires_at_str,
            audit_ids,
            trust_id,
            federated_group_ids,
            identity_provider_id,
            protocol_id,
            access_token_id,
            app_cred_id,
            thumbprint,
        )


_PAYLOAD_CLASSES = [
    UnscopedPayload,
    DomainScopedPayload,
    ProjectScopedPayload,
    TrustScopedPayload,
    FederatedUnscopedPayload,
    FederatedProjectScopedPayload,
    FederatedDomainScopedPayload,
    OauthScopedPayload,
    SystemScopedPayload,
    ApplicationCredentialScopedPayload,
    Oauth2CredentialsScopedPayload,
]



# Copyright 2022 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from keystone.conf import utils

oauth2_authn_methods = cfg.ListOpt(
    'oauth2_authn_methods',
    default=['tls_client_auth', 'client_secret_basic'],
    help=utils.fmt(
        """
The OAuth2.0 authentication method supported by the system when user obtains
an access token through the OAuth2.0 token endpoint. This option can be set to
certificate or secret. If the option is not set, the default value is
certificate. When the option is set to secret, the OAuth2.0 token endpoint
uses client_secret_basic method for authentication, otherwise tls_client_auth
method is used for authentication.
"""
    ),
)

oauth2_cert_dn_mapping_id = cfg.StrOpt(
    'oauth2_cert_dn_mapping_id',
    default='oauth2_mapping',
    help=utils.fmt(
        """
Used to define the mapping rule id. When not set, the mapping rule id is
oauth2_mapping.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [oauth2_authn_methods, oauth2_cert_dn_mapping_id]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Copyright 2018 Catalyst Cloud Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import http.client

import flask
from oslo_serialization import jsonutils

from keystone.common import authorization
from keystone.common import provider_api
from keystone import exception

PROVIDERS = provider_api.ProviderAPIs


def extract_receipt(auth_context):
    receipt_id = flask.request.headers.get(
        authorization.AUTH_RECEIPT_HEADER, None
    )
    if receipt_id:
        receipt = PROVIDERS.receipt_provider_api.validate_receipt(receipt_id)

        if auth_context['user_id'] != receipt.user_id:
            raise exception.ReceiptNotFound(
                "AuthContext user_id: {} does not match "
                "user_id for supplied auth receipt: {}".format(
                    auth_context['user_id'], receipt.user_id
                ),
                receipt_id=receipt_id,
            )
    else:
        receipt = None
    return receipt


def _render_receipt_response_from_model(receipt):
    receipt_reference = {
        'receipt': {
            'methods': receipt.methods,
            'user': {
                'id': receipt.user['id'],
                'name': receipt.user['name'],
                'domain': {
                    'id': receipt.user_domain['id'],
                    'name': receipt.user_domain['name'],
                },
            },
            'expires_at': receipt.expires_at,
            'issued_at': receipt.issued_at,
        },
        'required_auth_methods': receipt.required_methods,
    }
    return receipt_reference


def build_receipt(mfa_error):
    receipt = PROVIDERS.receipt_provider_api.issue_receipt(
        mfa_error.user_id, mfa_error.methods
    )
    resp_data = _render_receipt_response_from_model(receipt)
    resp_body = jsonutils.dumps(resp_data)
    response = flask.make_response(resp_body, http.client.UNAUTHORIZED)
    response.headers[authorization.AUTH_RECEIPT_HEADER] = receipt.id
    response.headers['Content-Type'] = 'application/json'
    return response



# Copyright 2015 Hewlett-Packard
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib

from oslo_log import log

from keystone.auth import core
from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone.federation import constants as federation_constants
from keystone.federation import utils
from keystone.i18n import _

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)


class TokenlessAuthHelper(provider_api.ProviderAPIMixin):
    def __init__(self, env):
        """A init class for TokenlessAuthHelper.

        :param env: The HTTP request environment that should contain
            client certificate attributes. These attributes should match
            with what the mapping defines. Or a user cannot be mapped and
            results un-authenticated. The following examples are for the
            attributes that reference to the client certificate's Subject's
            Common Name and Organization:
            SSL_CLIENT_S_DN_CN, SSL_CLIENT_S_DN_O
        :type env: dict
        """
        self.env = env

    def _build_scope_info(self):
        """Build the token request scope based on the headers.

        :returns: scope data
        :rtype: dict
        """
        project_id = self.env.get('HTTP_X_PROJECT_ID')
        project_name = self.env.get('HTTP_X_PROJECT_NAME')
        project_domain_id = self.env.get('HTTP_X_PROJECT_DOMAIN_ID')
        project_domain_name = self.env.get('HTTP_X_PROJECT_DOMAIN_NAME')
        domain_id = self.env.get('HTTP_X_DOMAIN_ID')
        domain_name = self.env.get('HTTP_X_DOMAIN_NAME')

        scope = {}
        if project_id:
            scope['project'] = {'id': project_id}
        elif project_name:
            scope['project'] = {'name': project_name}
            if project_domain_id:
                scope['project']['domain'] = {'id': project_domain_id}
            elif project_domain_name:
                scope['project']['domain'] = {'name': project_domain_name}
            else:
                msg = _(
                    'Neither Project Domain ID nor Project Domain Name '
                    'was provided.'
                )
                raise exception.ValidationError(msg)
        elif domain_id:
            scope['domain'] = {'id': domain_id}
        elif domain_name:
            scope['domain'] = {'name': domain_name}
        else:
            raise exception.ValidationError(
                attribute='project or domain', target='scope'
            )
        return scope

    def get_scope(self):
        auth = {}
        # NOTE(chioleong): Auth methods here are insignificant because
        # we only care about using auth.controllers.AuthInfo
        # to validate the scope information. Therefore,
        # we don't provide any identity.
        auth['scope'] = self._build_scope_info()

        # NOTE(chioleong): We'll let AuthInfo validate the scope for us
        auth_info = core.AuthInfo.create(auth, scope_only=True)
        return auth_info.get_scope()

    def get_mapped_user(self, project_id=None, domain_id=None):
        """Map client certificate to an existing user.

        If user is ephemeral, there is no validation on the user himself;
        however it will be mapped to a corresponding group(s) and the scope
        of this ephemeral user is the same as what is assigned to the group.

        :param project_id:  Project scope of the mapped user.
        :param domain_id: Domain scope of the mapped user.
        :returns: A dictionary that contains the keys, such as
            user_id, user_name, domain_id, domain_name
        :rtype: dict
        """
        idp_id = self._build_idp_id()
        LOG.debug(
            'The IdP Id %s and protocol Id %s are used to look up '
            'the mapping.',
            idp_id,
            CONF.tokenless_auth.protocol,
        )

        mapped_properties, mapping_id = self.federation_api.evaluate(
            idp_id, CONF.tokenless_auth.protocol, self.env
        )

        user = mapped_properties.get('user', {})
        user_id = user.get('id')
        user_name = user.get('name')
        user_type = user.get('type')
        if user.get('domain') is not None:
            user_domain_id = user.get('domain').get('id')
            user_domain_name = user.get('domain').get('name')
        else:
            user_domain_id = None
            user_domain_name = None

        # if user is ephemeral type, we don't care if the user exists
        # or not, but just care if the mapped group(s) is valid.
        if user_type == utils.UserType.EPHEMERAL:
            user_ref = {'type': utils.UserType.EPHEMERAL}
            group_ids = mapped_properties['group_ids']
            utils.validate_mapped_group_ids(
                group_ids, mapping_id, self.identity_api
            )
            group_ids.extend(
                utils.transform_to_group_ids(
                    mapped_properties['group_names'],
                    mapping_id,
                    self.identity_api,
                    self.resource_api,
                )
            )
            roles = self.assignment_api.get_roles_for_groups(
                group_ids, project_id, domain_id
            )
            if roles is not None:
                role_names = [role['name'] for role in roles]
                user_ref['roles'] = role_names
            user_ref['group_ids'] = list(group_ids)
            user_ref[federation_constants.IDENTITY_PROVIDER] = idp_id
            user_ref[federation_constants.PROTOCOL] = (
                CONF.tokenless_auth.protocol
            )
            return user_ref

        if user_id:
            user_ref = self.identity_api.get_user(user_id)
        elif user_name and (user_domain_name or user_domain_id):
            if user_domain_name:
                user_domain = self.resource_api.get_domain_by_name(
                    user_domain_name
                )
                self.resource_api.assert_domain_enabled(
                    user_domain['id'], user_domain
                )
                user_domain_id = user_domain['id']
            user_ref = self.identity_api.get_user_by_name(
                user_name, user_domain_id
            )
        else:
            msg = _(
                'User auth cannot be built due to missing either '
                'user id, or user name with domain id, or user name '
                'with domain name.'
            )
            raise exception.ValidationError(msg)
        self.identity_api.assert_user_enabled(
            user_id=user_ref['id'], user=user_ref
        )
        user_ref['type'] = utils.UserType.LOCAL
        return user_ref

    def _build_idp_id(self):
        """Build the IdP name from the given config option issuer_attribute.

        The default issuer attribute SSL_CLIENT_I_DN in the environment is
        built with the following formula -

        base64_idp = sha1(env['SSL_CLIENT_I_DN'])

        :returns: base64_idp like the above example
        :rtype: str
        """
        idp = self.env.get(CONF.tokenless_auth.issuer_attribute)
        if idp is None:
            raise exception.TokenlessAuthConfigError(
                issuer_attribute=CONF.tokenless_auth.issuer_attribute
            )

        hashed_idp = hashlib.sha256(idp.encode('utf-8'))
        return hashed_idp.hexdigest()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib
import uuid

import fixtures
from oslo_log import log

from keystone.common import fernet_utils
from keystone.credential.providers import fernet as credential_fernet
from keystone.tests import unit
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import database


class TestFernetCredentialProvider(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.provider = credential_fernet.Provider()
        self.useFixture(database.Database())
        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'credential',
                credential_fernet.MAX_ACTIVE_KEYS,
            )
        )

    def test_valid_data_encryption(self):
        blob = uuid.uuid4().hex
        encrypted_blob, primary_key_hash = self.provider.encrypt(blob)
        decrypted_blob = self.provider.decrypt(encrypted_blob)

        self.assertNotEqual(blob, encrypted_blob)
        self.assertEqual(blob, decrypted_blob)
        self.assertIsNotNone(primary_key_hash)


class TestFernetCredentialProviderWithNullKey(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.provider = credential_fernet.Provider()
        self.useFixture(database.Database())
        # Only do this to set the key_repository location in configuration. To
        # test the null key path, we need to make it so that the key repository
        # doesn't actually exist. If you're running the tests locally and have
        # bootstrapped a credential key repository in
        # `/etc/keystone/credential-keys` this will fail unless we override the
        # default.
        self.config_fixture.config(
            group='credential',
            key_repository=self.useFixture(fixtures.TempDir()).path,
        )

    def test_encryption_with_null_key(self):
        null_key = fernet_utils.NULL_KEY
        # NOTE(lhinds) This is marked as #nosec since bandit will see SHA1
        # which is marked insecure. Keystone uses SHA1 in this case as part of
        # HMAC-SHA1 which is currently not insecure but will still get
        # caught when scanning with bandit.
        null_key_hash = hashlib.sha1(null_key).hexdigest()  # nosec

        blob = uuid.uuid4().hex
        encrypted_blob, primary_key_hash = self.provider.encrypt(blob)
        self.assertEqual(null_key_hash, primary_key_hash)
        self.assertNotEqual(blob, encrypted_blob)

        decrypted_blob = self.provider.decrypt(encrypted_blob)
        self.assertEqual(blob, decrypted_blob)

    def test_warning_is_logged_when_encrypting_with_null_key(self):
        blob = uuid.uuid4().hex
        logging_fixture = self.useFixture(fixtures.FakeLogger(level=log.DEBUG))
        expected_output = (
            'Encrypting credentials with the null key. Please properly '
            'encrypt credentials using `keystone-manage credential_setup`, '
            '`keystone-manage credential_migrate`, and `keystone-manage '
            'credential_rotate`'
        )
        encrypted_blob, primary_key_hash = self.provider.encrypt(blob)
        self.assertIn(expected_output, logging_fixture.output)
